import contextlib
import time

import penn


###############################################################################
# Profiling utilities
###############################################################################


class Context:
    """Context manager timer"""

    def __init__(self):
        self.reset()

    def __call__(self):
        """Retrieve timer results"""
        return {name: sum(times) for name, times in self.history.items()}

    def __enter__(self):
        """Start the timer"""
        self.start = time.time()

    def __exit__(self, *_):
        """Stop the timer"""
        elapsed = time.time() - self.start

        # Add to timer history
        if self.name not in self.history:
            self.history[self.name] = [elapsed]
        else:
            self.history[self.name].append(elapsed)

    def reset(self):
        """Reset the timer"""
        self.history = {}
        self.start = 0.
        self.name = None


@contextlib.contextmanager
def timer(name):
    """Wrapper to handle context changes of global timer"""
    # Don't continue if we aren't benchmarking
    if not penn.BENCHMARK:
        yield
        return

    penn.TIMER.name = name
    with penn.TIMER:
        yield
    penn.TIMER.name = None



import math

import torch

import penn


###############################################################################
# Batch sampler
###############################################################################


def sampler(dataset, partition):
    """Create batch index sampler"""
    # Get sampler indices
    indices = (
        dataset.voiced_indices() if penn.VOICED_ONLY and partition == 'train'
        else list(range(len(dataset))))

    # Maybe use distributed sampler for training
    if partition == 'train':
        return (
            DistributedSampler(indices) if torch.distributed.is_initialized()
            else Sampler(indices))

    # Possibly deterministic random sampler for validation
    elif partition == 'valid':
        return Sampler(indices)

    # Sample test data sequentially
    elif partition == 'test':
        return torch.utils.data.SequentialSampler(dataset)

    else:
        raise ValueError(f'Partition {partition} is not implemented')


###############################################################################
# Custom samplers
###############################################################################


class Sampler:

    def __init__(self, indices):
        self.indices = indices
        self.epoch = 0

    def __iter__(self):
        generator = torch.Generator()
        generator.manual_seed(penn.RANDOM_SEED + self.epoch)
        for i in torch.randperm(len(self.indices), generator=generator):
            yield self.indices[i]

    def __len__(self):
        return len(self.indices)

    def set_epoch(self, epoch):
        self.epoch = epoch


class DistributedSampler:

    def __init__(self, indices):
        self.indices = indices
        self.epoch = 0
        self.rank = torch.distributed.get_rank()
        self.num_replicas = torch.distributed.get_world_size()
        self.num_samples = math.ceil(len(self.indices) / self.num_replicas)
        self.total_size = self.num_samples * self.num_replicas

    def __iter__(self):
        # Deterministic shuffling based on epoch
        generator = torch.Generator()
        generator.manual_seed(penn.RANDOM_SEED + self.epoch)
        indices = [
            self.indices[i] for i in
            torch.randperm(len(self.indices), generator=generator)]

        # Add extra samples to make it evenly divisible
        padding = self.total_size - len(indices)
        if padding <= len(indices):
            indices += indices[:padding]
        else:
            indices += (
                indices * math.ceil(padding / len(indices)))[:padding]

        # Subsample
        return iter(indices[self.rank:self.total_size:self.num_replicas])

    def __len__(self):
        return self.num_samples

    def set_epoch(self, epoch):
        self.epoch = epoch



from .core import *
from .crepe import Crepe
from .deepf0 import Deepf0
from .fcnf0 import Fcnf0

import penn


def Model(name=penn.MODEL):
    """Create a model"""
    if name == 'crepe':
        return Crepe()
    if name == 'deepf0':
        return Deepf0()
    if name == 'fcnf0':
        return Fcnf0()
    raise ValueError(f'Model {name} is not defined')



import argparse
from pathlib import Path

import penn


###############################################################################
# Evaluate pitch and periodicity estimation
###############################################################################


def parse_args():
    """Parse command-line arguments"""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--datasets',
        nargs='+',
        default=penn.EVALUATION_DATASETS,
        help='The datasets to evaluate on')
    parser.add_argument(
        '--checkpoint',
        type=Path,
        default=penn.DEFAULT_CHECKPOINT,
        help='The checkpoint file to evaluate')
    parser.add_argument(
        '--gpu',
        type=int,
        help='The index of the GPU to use for evaluation')

    return parser.parse_known_args()[0]


penn.evaluate.datasets(**vars(parse_args()))



import torch

import penn


###############################################################################
# Voiced/unvoiced
###############################################################################


def interpolate(pitch, periodicity, value=penn.DEFAULT_VOICING_THRESHOLD):
    """Fill unvoiced regions via linear interpolation"""
    # Threshold periodicity
    voiced = threshold(periodicity, value)

    # Handle no voiced frames
    if not voiced.any():
        return pitch

    # Pitch is linear in base-2 log-space
    pitch = torch.log2(pitch)

    # Interpolate
    pitch[~voiced] = penn.interpolate(
        torch.where(~voiced[0])[0][None],
        torch.where(voiced[0])[0][None],
        pitch[voiced][None])

    return 2 ** pitch


def threshold(periodicity, value=penn.DEFAULT_VOICING_THRESHOLD):
    """Threshold periodicity to produce voiced/unvoiced classifications"""
    return periodicity > value



import contextlib
import os
from pathlib import Path
from typing import List, Optional, Tuple

import torch
import torchaudio
import tqdm

import penn


###############################################################################
# Pitch and periodicity estimation
###############################################################################


def from_audio(
        audio: torch.Tensor,
        sample_rate: int,
        hopsize: float = penn.HOPSIZE_SECONDS,
        fmin: float = penn.FMIN,
        fmax: float = penn.FMAX,
        checkpoint: Path = penn.DEFAULT_CHECKPOINT,
        batch_size: Optional[int] = None,
        pad: bool = False,
        interp_unvoiced_at: Optional[float] = penn.DEFAULT_VOICING_THRESHOLD,
        gpu: Optional[int] = None) -> Tuple[torch.Tensor, torch.Tensor]:
    """Perform pitch and periodicity estimation

    Args:
        audio: The audio to extract pitch and periodicity from
        sample_rate: The audio sample rate
        hopsize: The hopsize in seconds
        fmin: The minimum allowable frequency in Hz
        fmax: The maximum allowable frequency in Hz
        checkpoint: The checkpoint file
        batch_size: The number of frames per batch
        pad: If true, centers frames at hopsize / 2, 3 * hopsize / 2, 5 * ...
        interp_unvoiced_at: Specifies voicing threshold for interpolation
        gpu: The index of the gpu to run inference on

    Returns:
        pitch: torch.tensor(
            shape=(1, int(samples // penn.seconds_to_sample(hopsize))))
        periodicity: torch.tensor(
            shape=(1, int(samples // penn.seconds_to_sample(hopsize))))
    """
    pitch, periodicity = [], []

    # Preprocess audio
    for frames, _ in preprocess(audio, sample_rate, hopsize, batch_size, pad):

        # Copy to device
        with penn.time.timer('copy-to'):
            frames = frames.to('cpu' if gpu is None else f'cuda:{gpu}')

        # Infer
        logits = infer(frames, checkpoint).detach()

        # Postprocess
        with penn.time.timer('postprocess'):
            result = postprocess(logits, fmin, fmax)
            pitch.append(result[1])
            periodicity.append(result[2])

    # Concatenate results
    pitch, periodicity = torch.cat(pitch, 1), torch.cat(periodicity, 1)

    # Maybe interpolate unvoiced regions
    if interp_unvoiced_at is not None:
        pitch = penn.voicing.interpolate(
            pitch,
            periodicity,
            interp_unvoiced_at)

    return pitch, periodicity


def from_file(
        file: Path,
        hopsize: float = penn.HOPSIZE_SECONDS,
        fmin: float = penn.FMIN,
        fmax: float = penn.FMAX,
        checkpoint: Path = penn.DEFAULT_CHECKPOINT,
        batch_size: Optional[int] = None,
        pad: bool = False,
        interp_unvoiced_at: Optional[float] = penn.DEFAULT_VOICING_THRESHOLD,
        gpu: Optional[int] = None) -> Tuple[torch.Tensor, torch.Tensor]:
    """Perform pitch and periodicity estimation from audio on disk

    Args:
        file: The audio file
        hopsize: The hopsize in seconds
        fmin: The minimum allowable frequency in Hz
        fmax: The maximum allowable frequency in Hz
        checkpoint: The checkpoint file
        batch_size: The number of frames per batch
        pad: If true, centers frames at hopsize / 2, 3 * hopsize / 2, 5 * ...
        interp_unvoiced_at: Specifies voicing threshold for interpolation
        gpu: The index of the gpu to run inference on

    Returns:
        pitch: torch.tensor(shape=(1, int(samples // hopsize)))
        periodicity: torch.tensor(shape=(1, int(samples // hopsize)))
    """
    # Load audio
    with penn.time.timer('load'):
        audio = penn.load.audio(file)

    # Inference
    return from_audio(
        audio,
        penn.SAMPLE_RATE,
        hopsize,
        fmin,
        fmax,
        checkpoint,
        batch_size,
        pad,
        interp_unvoiced_at,
        gpu)


def from_file_to_file(
        file: Path,
        output_prefix: Optional[Path] = None,
        hopsize: float = penn.HOPSIZE_SECONDS,
        fmin: float = penn.FMIN,
        fmax: float = penn.FMAX,
        checkpoint: Path = penn.DEFAULT_CHECKPOINT,
        batch_size: Optional[int] = None,
        pad: bool = False,
        interp_unvoiced_at: Optional[float] = penn.DEFAULT_VOICING_THRESHOLD,
        gpu: Optional[int] = None) -> None:
    """Perform pitch and periodicity estimation from audio on disk and save

    Args:
        file: The audio file
        output_prefix: The file to save pitch and periodicity without extension
        hopsize: The hopsize in seconds
        fmin: The minimum allowable frequency in Hz
        fmax: The maximum allowable frequency in Hz
        checkpoint: The checkpoint file
        batch_size: The number of frames per batch
        pad: If true, centers frames at hopsize / 2, 3 * hopsize / 2, 5 * ...
        interp_unvoiced_at: Specifies voicing threshold for interpolation
        gpu: The index of the gpu to run inference on
    """
    # Inference
    pitch, periodicity = from_file(
        file,
        hopsize,
        fmin,
        fmax,
        checkpoint,
        batch_size,
        pad,
        interp_unvoiced_at,
        gpu)

    # Move to cpu
    with penn.time.timer('copy-from'):
        pitch, periodicity = pitch.cpu(), periodicity.cpu()

    # Save to disk
    with penn.time.timer('save'):

        # Maybe use same filename with new extension
        if output_prefix is None:
            output_prefix = file.parent / file.stem

        # Save
        torch.save(pitch, f'{output_prefix}-pitch.pt')
        torch.save(periodicity, f'{output_prefix}-periodicity.pt')


def from_files_to_files(
    files: List[Path],
    output_prefixes: Optional[List[Path]] = None,
    hopsize: float = penn.HOPSIZE_SECONDS,
    fmin: float = penn.FMIN,
    fmax: float = penn.FMAX,
    checkpoint: Path = penn.DEFAULT_CHECKPOINT,
    batch_size: Optional[int] = None,
    pad: bool = False,
    interp_unvoiced_at: Optional[float] = penn.DEFAULT_VOICING_THRESHOLD,
    gpu: Optional[int] = None) -> None:
    """Perform pitch and periodicity estimation from files on disk and save

    Args:
        files: The audio files
        output_prefixes: Files to save pitch and periodicity without extension
        hopsize: The hopsize in seconds
        fmin: The minimum allowable frequency in Hz
        fmax: The maximum allowable frequency in Hz
        checkpoint: The checkpoint file
        batch_size: The number of frames per batch
        pad: If true, centers frames at hopsize / 2, 3 * hopsize / 2, 5 * ...
        interp_unvoiced_at: Specifies voicing threshold for interpolation
        gpu: The index of the gpu to run inference on
    """
    # Maybe use default output filenames
    if output_prefixes is None:
        output_prefixes = len(files) * [None]

    # Iterate over files
    for file, output_prefix in iterator(
        zip(files, output_prefixes),
        f'{penn.CONFIG}',
        total=len(files)):

        # Infer
        from_file_to_file(
            file,
            output_prefix,
            hopsize,
            fmin,
            fmax,
            checkpoint,
            batch_size,
            pad,
            interp_unvoiced_at,
            gpu)


###############################################################################
# Inference pipeline stages
###############################################################################


def infer(frames, checkpoint=penn.DEFAULT_CHECKPOINT):
    """Forward pass through the model"""
    # Time model loading
    with penn.time.timer('model'):

        # Load and cache model
        if (
            not hasattr(infer, 'model') or
            infer.checkpoint != checkpoint or
            infer.device_type != frames.device.type
        ):

            # Maybe initialize model
            model = penn.Model()

            # Load from disk
            infer.model, *_ = penn.checkpoint.load(checkpoint, model)
            infer.checkpoint = checkpoint
            infer.device_type = frames.device.type

            # Move model to correct device (no-op if devices are the same)
            infer.model = infer.model.to(frames.device)

    # Time inference
    with penn.time.timer('infer'):

        # Prepare model for inference
        with inference_context(infer.model):

            # Infer
            logits = infer.model(frames)

        # If we're benchmarking, make sure inference finishes within timer
        if penn.BENCHMARK and logits.device.type == 'cuda':
            torch.cuda.synchronize(logits.device)

        return logits


def postprocess(logits, fmin=penn.FMIN, fmax=penn.FMAX):
    """Convert model output to pitch and periodicity"""
    # Turn off gradients
    with torch.no_grad():

        # Convert frequency range to pitch bin range
        minidx = penn.convert.frequency_to_bins(torch.tensor(fmin))
        maxidx = penn.convert.frequency_to_bins(
            torch.tensor(fmax),
            torch.ceil)

        # Remove frequencies outside of allowable range
        logits[:, :minidx] = -float('inf')
        logits[:, maxidx:] = -float('inf')

        # Decode pitch from logits
        if penn.DECODER == 'argmax':
            bins, pitch = penn.decode.argmax(logits)
        elif penn.DECODER.startswith('viterbi'):
            bins, pitch = penn.decode.viterbi(logits)
        elif penn.DECODER == 'locally_normal':
            bins, pitch = penn.decode.locally_normal(logits)
        else:
            raise ValueError(f'Decoder method {penn.DECODER} is not defined')

        # Decode periodicity from logits
        if penn.PERIODICITY == 'entropy':
            periodicity = penn.periodicity.entropy(logits)
        elif penn.PERIODICITY == 'max':
            periodicity = penn.periodicity.max(logits)
        elif penn.PERIODICITY == 'sum':
            periodicity = penn.periodicity.sum(logits)
        else:
            raise ValueError(
                f'Periodicity method {penn.PERIODICITY} is not defined')

        return bins.T, pitch.T, periodicity.T


def preprocess(
    audio,
    sample_rate,
    hopsize=penn.HOPSIZE_SECONDS,
    batch_size=None,
    pad=False):
    """Convert audio to model input"""
    # Convert hopsize to samples
    hopsize = int(penn.convert.seconds_to_samples(hopsize))

    # Resample
    if sample_rate != penn.SAMPLE_RATE:
        audio = resample(audio, sample_rate)

    # Maybe pad audio
    padding = int((penn.WINDOW_SIZE - hopsize) / 2)
    if pad:
        audio = torch.nn.functional.pad(audio, (padding, padding))

    # Get total number of frames
    total_frames = int((audio.shape[-1] - 2 * padding) / hopsize)

    # Default to running all frames in a single batch
    batch_size = total_frames if batch_size is None else batch_size

    # Generate batches
    for i in range(0, total_frames, batch_size):

        # Size of this batch
        batch = min(total_frames - i, batch_size)

        # Batch indices
        start = i * hopsize
        end = start + int((batch - 1) * hopsize) + penn.WINDOW_SIZE
        end = min(end, audio.shape[-1])
        batch_audio = audio[:, start:end]

        # Maybe pad to a single frame
        if end - start < penn.WINDOW_SIZE:
            padding = penn.WINDOW_SIZE - (end - start)

            # Handle multiple of hopsize
            remainder = (end - start) % penn.HOPSIZE
            if remainder:
                padding += end - start - penn.HOPSIZE

            # Pad
            batch_audio = torch.nn.functional.pad(batch_audio, (0, padding))

        # Slice and chunk audio
        frames = torch.nn.functional.unfold(
            batch_audio[:, None, None],
            kernel_size=(1, penn.WINDOW_SIZE),
            stride=(1, hopsize)).permute(2, 0, 1)

        yield frames, batch


###############################################################################
# Utilities
###############################################################################


def cents(a, b):
    """Compute pitch difference in cents"""
    return penn.OCTAVE * torch.log2(a / b)


@contextlib.contextmanager
def chdir(directory):
    """Context manager for changing the current working directory"""
    curr_dir = os.getcwd()
    try:
        os.chdir(directory)
        yield
    finally:
        os.chdir(curr_dir)


@contextlib.contextmanager
def inference_context(model):
    device_type = next(model.parameters()).device.type

    # Prepare model for evaluation
    model.eval()

    # Turn off gradient computation
    with torch.no_grad():

        # Automatic mixed precision on GPU
        if device_type == 'cuda':
            with torch.autocast(device_type):
                yield

        else:
            yield

    # Prepare model for training
    model.train()


def interpolate(x, xp, fp):
    """1D linear interpolation for monotonically increasing sample points"""
    # Handle edge cases
    if xp.shape[-1] == 0:
        return x
    if xp.shape[-1] == 1:
        return torch.full(
            x.shape,
            fp.squeeze(),
            device=fp.device,
            dtype=fp.dtype)

    # Get slope and intercept using right-side first-differences
    m = (fp[:, 1:] - fp[:, :-1]) / (xp[:, 1:] - xp[:, :-1])
    b = fp[:, :-1] - (m.mul(xp[:, :-1]))

    # Get indices to sample slope and intercept
    indicies = torch.sum(torch.ge(x[:, :, None], xp[:, None, :]), -1) - 1
    indicies = torch.clamp(indicies, 0, m.shape[-1] - 1)
    line_idx = torch.linspace(
        0,
        indicies.shape[0],
        1,
        device=indicies.device).to(torch.long).expand(indicies.shape)

    # Interpolate
    return m[line_idx, indicies].mul(x) + b[line_idx, indicies]


def iterator(iterable, message, initial=0, total=None):
    """Create a tqdm iterator"""
    total = len(iterable) if total is None else total
    return tqdm.tqdm(
        iterable,
        desc=message,
        dynamic_ncols=True,
        initial=initial,
        total=total)


def normalize(frames):
    """Normalize audio frames to have mean zero and std dev one"""
    # Mean-center
    frames -= frames.mean(dim=2, keepdim=True)

    # Scale
    frames /= torch.max(
        torch.tensor(1e-10, device=frames.device),
        frames.std(dim=2, keepdim=True))

    return frames


def resample(audio, sample_rate, target_rate=penn.SAMPLE_RATE):
    """Perform audio resampling"""
    if sample_rate == target_rate:
        return audio
    resampler = torchaudio.transforms.Resample(sample_rate, target_rate)
    resampler = resampler.to(audio.device)
    return resampler(audio)



import torch

import penn


###############################################################################
# Pitch conversions
###############################################################################


def bins_to_cents(bins):
    """Converts pitch bins to cents"""
    return penn.CENTS_PER_BIN * bins


def bins_to_frequency(bins):
    """Converts pitch bins to frequency in Hz"""
    return cents_to_frequency(bins_to_cents(bins))


def cents_to_bins(cents, quantize_fn=torch.floor):
    """Converts cents to pitch bins"""
    bins = quantize_fn(cents / penn.CENTS_PER_BIN).long()
    bins[bins < 0] = 0
    bins[bins >= penn.PITCH_BINS] = penn.PITCH_BINS - 1
    return bins


def cents_to_frequency(cents):
    """Converts cents to frequency in Hz"""
    return penn.FMIN * 2 ** (cents / penn.OCTAVE)


def frequency_to_bins(frequency, quantize_fn=torch.floor):
    """Convert frequency in Hz to pitch bins"""
    return cents_to_bins(frequency_to_cents(frequency), quantize_fn)


def frequency_to_cents(frequency):
    """Convert frequency in Hz to cents"""
    return penn.OCTAVE * torch.log2(frequency / penn.FMIN)


def frequency_to_samples(frequency, sample_rate=penn.SAMPLE_RATE):
    """Convert frequency in Hz to number of samples per period"""
    return sample_rate / frequency


###############################################################################
# Time conversions
###############################################################################


def frames_to_samples(frames):
    """Convert number of frames to samples"""
    return frames * penn.HOPSIZE


def frames_to_seconds(frames):
    """Convert number of frames to seconds"""
    return frames * penn.HOPSIZE_SECONDS


def seconds_to_frames(seconds):
    """Convert seconds to number of frames"""
    return samples_to_frames(seconds_to_samples(seconds))


def seconds_to_samples(seconds):
    """Convert seconds to number of samples"""
    return seconds * penn.SAMPLE_RATE


def samples_to_frames(samples):
    """Convert samples to number of frames"""
    return samples // penn.HOPSIZE


def samples_to_seconds(samples):
    """Convert number of samples to seconds"""
    return samples / penn.SAMPLE_RATE



import argparse
from pathlib import Path

import penn


###############################################################################
# Create figure
###############################################################################


def parse_args():
    """Parse command-line arguments"""
    parser = argparse.ArgumentParser(description='Create density figure')
    parser.add_argument(
        '--datasets',
        nargs='+',
        required=True,
        help='Datasets to use for density figure')
    parser.add_argument(
        '--output_file',
        required=True,
        type=Path,
        help='The jpg file to save the plot')
    parser.add_argument(
        '--checkpoint',
        type=Path,
        default=penn.DEFAULT_CHECKPOINT,
        help='The checkpoint file to use for inference')
    parser.add_argument(
        '--gpu',
        type=int,
        help='The index of the GPU to use for inference')
    return parser.parse_known_args()[0]

penn.plot.density.to_file(**vars(parse_args()))



import bisect

import numpy as np
import torch

import penn


###############################################################################
# Dataset
###############################################################################


class Dataset(torch.utils.data.Dataset):
    """PyTorch dataset

    Arguments
        names - list[str]
            The names of datasets to load from
        partition - string
            The name of the data partition
    """

    def __init__(self, names, partition, hparam_search=False):
        self.partition = partition
        self.hparam_search = hparam_search
        self.datasets = [Metadata(name, partition) for name in names]

    def __getitem__(self, index):
        if (
            self.partition == 'test' or
            (self.partition == 'valid' and self.hparam_search)
        ):
            return self.load_inference(index)
        return self.load_training(index)

    def __len__(self):
        """Length of the dataset"""
        if (
            self.partition == 'test' or
            (self.partition == 'valid' and self.hparam_search)
        ):
            return sum(len(dataset.files) for dataset in self.datasets)
        return sum(dataset.total for dataset in self.datasets)

    def load_inference(self, index):
        """Load item for inference"""
        # Get dataset to query
        i = 0
        dataset = self.datasets[i]
        upper_bound = len(dataset.files)
        while index >= upper_bound:
            i += 1
            dataset = self.datasets[i]
            upper_bound += len(dataset.files)

        # Get index into dataset
        index -= (upper_bound - len(dataset.files))

        # Get stem
        stem = dataset.stems[index]

        # Load from cache
        directory = penn.CACHE_DIR / dataset.name
        audio = np.load(directory / f'{stem}-audio.npy')
        pitch = np.load(directory / f'{stem}-pitch.npy')
        voiced = np.load(directory / f'{stem}-voiced.npy')

        # Convert to torch
        audio = torch.from_numpy(audio)[None]
        pitch = torch.from_numpy(pitch)
        voiced = torch.from_numpy(voiced)

        # Convert to pitch bin categories
        bins = penn.convert.frequency_to_bins(pitch)

        # Set unvoiced bins to random values
        bins = torch.where(
            ~voiced,
            torch.randint(0, penn.PITCH_BINS, bins.shape, dtype=torch.long),
            bins)

        return audio, bins, pitch, voiced, stem

    def load_training(self, index):
        """Load item for training"""
        # Get dataset to query
        i = 0
        dataset = self.datasets[i]
        upper_bound = dataset.total
        while index >= upper_bound:
            i += 1
            dataset = self.datasets[i]
            upper_bound += dataset.total

        # Get index into dataset
        index -= (upper_bound - dataset.total)

        # Get stem
        stem_index = bisect.bisect(dataset.offsets, index)
        stem = dataset.stems[stem_index]

        # Get start and end frames
        start = \
            index - (0 if stem_index == 0 else dataset.offsets[stem_index - 1])
        end = start + penn.NUM_TRAINING_FRAMES

        # Get start and end samples
        start_sample = \
            penn.convert.frames_to_samples(start) - \
            (penn.WINDOW_SIZE - penn.HOPSIZE) // 2
        end_sample = start_sample + penn.NUM_TRAINING_SAMPLES

        # Load from cache
        directory = penn.CACHE_DIR / dataset.name
        waveform = np.load(directory / f'{stem}-audio.npy', mmap_mode='r')
        pitch = np.load(directory / f'{stem}-pitch.npy', mmap_mode='r')
        voiced = np.load(directory / f'{stem}-voiced.npy', mmap_mode='r')

        # Slice audio
        if start_sample < 0:
            audio = torch.zeros(
                (penn.NUM_TRAINING_SAMPLES,),
                dtype=torch.float)
            audio[-start_sample:] = torch.from_numpy(
                waveform[:end_sample].copy())
        elif end_sample > len(waveform):
            audio = torch.zeros(
                (penn.NUM_TRAINING_SAMPLES,),
                dtype=torch.float)
            audio[:len(waveform) - end_sample] = torch.from_numpy(
                waveform[start_sample:].copy())
        else:
            audio = torch.from_numpy(
                waveform[start_sample:end_sample].copy())

        # Slice pitch and voicing
        pitch = torch.from_numpy(pitch[start:end].copy())
        voiced = torch.from_numpy(voiced[start:end].copy())

        # Convert to pitch bin categories
        bins = penn.convert.frequency_to_bins(pitch)

        # Set unvoiced bins to random values
        bins = torch.where(
            ~voiced,
            torch.randint(0, penn.PITCH_BINS, bins.shape, dtype=torch.long),
            bins)

        return audio[None], bins, pitch, voiced, stem

    def voiced_indices(self):
        """Retrieve the indices with voiced start frames"""
        offset = 0
        indices = []
        for dataset in self.datasets:
            indices += [index + offset for index in dataset.voiced_indices()]
            offset += dataset.total
        return indices


###############################################################################
# Metadata
###############################################################################


class Metadata:

    def __init__(self, name, partition):
        self.name = name
        self.stems = penn.load.partition(name)[partition]
        self.files = [
            penn.CACHE_DIR / name / f'{stem}-audio.npy'
            for stem in self.stems]

        # Get number of frames in each file
        self.frames = [
            penn.convert.samples_to_frames(len(np.load(file, mmap_mode='r')))
            for file in self.files]

        # We require all files to be at least as large as the analysis window
        assert all(frame >= penn.NUM_TRAINING_FRAMES for frame in self.frames)

        # Remove invalid center points
        self.frames = [
            frame - (penn.NUM_TRAINING_FRAMES - 1) for frame in self.frames]

        # Save frame offsets
        self.offsets = np.cumsum(self.frames)

        # Total number of valid start points
        self.total = self.offsets[-1]

    def voiced_indices(self):
        """Retrieve the indices with voiced start frames"""
        # Get voicing files
        files = [
            penn.CACHE_DIR / self.name / f'{stem}-voiced.npy'
            for stem in self.stems]

        offset = 0
        indices = []
        for file in files:

            # Load
            voiced = np.load(file)

            # Remove invalid center points
            if penn.NUM_TRAINING_FRAMES > 1:
                voiced = voiced[:-(penn.NUM_TRAINING_FRAMES - 1)]

            # Update
            indices.extend(list(voiced.nonzero()[0] + offset))
            offset += len(voiced)

        return indices



import functools
import multiprocessing as mp

import numpy as np
import torch

import penn


###############################################################################
# PYIN (from librosa)
###############################################################################


def from_audio(
    audio,
    sample_rate=penn.SAMPLE_RATE,
    hopsize=penn.HOPSIZE_SECONDS,
    fmin=penn.FMIN,
    fmax=penn.FMAX):
    """Estimate pitch and periodicity with pyin"""
    # Pad
    pad = int(
        penn.WINDOW_SIZE - penn.convert.seconds_to_samples(hopsize)) // 2
    audio = torch.nn.functional.pad(audio, (pad, pad))

    # Infer pitch bin probabilities
    with penn.time.timer('infer'):
        logits = infer(audio, sample_rate, hopsize, fmin, fmax)

    # Decode pitch and periodicity
    with penn.time.timer('postprocess'):
        return penn.postprocess(logits)[1:]


def from_file(
    file,
    hopsize=penn.HOPSIZE_SECONDS,
    fmin=penn.FMIN,
    fmax=penn.FMAX):
    """Estimate pitch and periodicity with pyin from audio on disk"""
    # Load
    with penn.time.timer('load'):
        audio = penn.load.audio(file)

    # Infer
    return from_audio(audio, penn.SAMPLE_RATE, hopsize, fmin, fmax)


def from_file_to_file(
    file,
    output_prefix=None,
    hopsize=penn.HOPSIZE_SECONDS,
    fmin=penn.FMIN,
    fmax=penn.FMAX):
    """Estimate pitch and periodicity with pyin and save to disk"""
    # Infer
    pitch, periodicity = from_file(file, hopsize, fmin, fmax)

    # Save to disk
    with penn.time.timer('save'):

        # Maybe use same filename with new extension
        if output_prefix is None:
            output_prefix = file.parent / file.stem

        # Save
        torch.save(pitch, f'{output_prefix}-pitch.pt')
        torch.save(periodicity, f'{output_prefix}-periodicity.pt')


def from_files_to_files(
    files,
    output_prefixes=None,
    hopsize=penn.HOPSIZE_SECONDS,
    fmin=penn.FMIN,
    fmax=penn.FMAX):
    """Estimate pitch and periodicity with pyin and save to disk"""
    pitch_fn = functools.partial(
        from_file_to_file,
        hopsize=hopsize,
        fmin=fmin,
        fmax=fmax)
    iterator = zip(files, output_prefixes)

    # Turn off multiprocessing for benchmarking
    if penn.BENCHMARK:
        iterator = penn.iterator(
            iterator,
            f'{penn.CONFIG}',
            total=len(files))
        for item in iterator:
            pitch_fn(*item)
    else:
        with mp.get_context('spawn').Pool() as pool:
            pool.starmap(pitch_fn, iterator)


###############################################################################
# Utilities
###############################################################################


def cumulative_mean_normalized_difference(frames, min_period, max_period):
    import librosa

    a = np.fft.rfft(frames, 2 * penn.WINDOW_SIZE, axis=-2)
    b = np.fft.rfft(
        frames[..., penn.WINDOW_SIZE:0:-1, :],
        2 * penn.WINDOW_SIZE,
        axis=-2)
    acf_frames = np.fft.irfft(
        a * b, 2 * penn.WINDOW_SIZE, axis=-2)[..., penn.WINDOW_SIZE:, :]
    acf_frames[np.abs(acf_frames) < 1e-6] = 0

    # Energy terms
    energy_frames = np.cumsum(frames ** 2, axis=-2)
    energy_frames = (
        energy_frames[..., penn.WINDOW_SIZE:, :] -
        energy_frames[..., :-penn.WINDOW_SIZE, :])
    energy_frames[np.abs(energy_frames) < 1e-6] = 0

    # Difference function
    yin_frames = energy_frames[..., :1, :] + energy_frames - 2 * acf_frames

    # Cumulative mean normalized difference function
    yin_numerator = yin_frames[..., min_period: max_period + 1, :]

    # Broadcast to have leading ones
    tau_range = librosa.util.expand_to(
        np.arange(1, max_period + 1), ndim=yin_frames.ndim, axes=-2)

    cumulative_mean = (
        np.cumsum(yin_frames[..., 1: max_period + 1, :], axis=-2) / tau_range)

    yin_denominator = cumulative_mean[..., min_period - 1: max_period, :]
    yin_frames = yin_numerator / \
        (yin_denominator + librosa.util.tiny(yin_denominator))

    return yin_frames


def infer(
    audio,
    sample_rate=penn.SAMPLE_RATE,
    hopsize=penn.HOPSIZE_SECONDS,
    fmin=penn.FMIN,
    fmax=penn.FMAX):
    hopsize = int(penn.convert.seconds_to_samples(hopsize))
    import scipy

    # Pad audio to center-align frames
    pad = penn.WINDOW_SIZE // 2
    padded = torch.nn.functional.pad(audio, (0, 2 * pad))

    # Slice and chunk audio
    frames = torch.nn.functional.unfold(
        padded[:, None, None],
        kernel_size=(1, 2 * penn.WINDOW_SIZE),
        stride=(1, penn.HOPSIZE))[0]

    # Calculate minimum and maximum periods
    min_period = max(int(np.floor(sample_rate / fmax)), 1)
    max_period = min(
        int(np.ceil(sample_rate / fmin)),
        penn.WINDOW_SIZE - 1)

    # Calculate cumulative mean normalized difference function
    yin_frames = cumulative_mean_normalized_difference(
        frames.numpy(),
        min_period,
        max_period)

    # Parabolic interpolation
    parabolic_shifts = parabolic_interpolation(yin_frames)

    # Find Yin candidates and probabilities.
    # The implementation here follows the official pYIN software which
    # differs from the method described in the paper.
    # 1. Define the prior over the thresholds.
    thresholds = np.linspace(0, 1, 100 + 1)
    beta_cdf = scipy.stats.beta.cdf(thresholds, 2, 18)
    beta_probs = np.diff(beta_cdf)

    def _helper(a, b):
        return pyin_helper(
            a,
            b,
            thresholds,
            2,
            beta_probs,
            .01,
            min_period,
            (penn.OCTAVE / 12) / penn.CENTS_PER_BIN)

    helper = np.vectorize(_helper, signature="(f,t),(k,t)->(1,d,t)")
    probs = helper(yin_frames, parabolic_shifts)
    probs = torch.from_numpy(probs)
    return torch.log(probs).permute(2, 1, 0)


def parabolic_interpolation(frames):
    """Piecewise parabolic interpolation for yin and pyin"""
    import librosa

    parabolic_shifts = np.zeros_like(frames)
    parabola_a = (
        frames[..., :-2, :] +
        frames[..., 2:, :] -
        2 * frames[..., 1:-1, :]
    ) / 2
    parabola_b = (frames[..., 2:, :] - frames[..., :-2, :]) / 2
    parabolic_shifts[..., 1:-1, :] = \
        -parabola_b / (2 * parabola_a + librosa.util.tiny(parabola_a))
    parabolic_shifts[np.abs(parabolic_shifts) > 1] = 0
    return parabolic_shifts


def pyin_helper(
        frames,
        parabolic_shifts,
        thresholds,
        boltzmann_parameter,
        beta_probs,
        no_trough_prob,
        min_period,
        n_bins_per_semitone):
    import librosa
    import scipy


    yin_probs = np.zeros_like(frames)

    for i, yin_frame in enumerate(frames.T):
        # 2. For each frame find the troughs.
        is_trough = librosa.util.localmin(yin_frame)

        is_trough[0] = yin_frame[0] < yin_frame[1]
        (trough_index,) = np.nonzero(is_trough)

        if len(trough_index) == 0:
            continue

        # 3. Find the troughs below each threshold.
        # these are the local minima of the frame, could get them directly
        # without the trough index
        trough_heights = yin_frame[trough_index]
        trough_thresholds = np.less.outer(trough_heights, thresholds[1:])

        # 4. Define the prior over the troughs.
        # Smaller periods are weighted more.
        trough_positions = np.cumsum(trough_thresholds, axis=0) - 1
        n_troughs = np.count_nonzero(trough_thresholds, axis=0)

        trough_prior = scipy.stats.boltzmann.pmf(
            trough_positions,
            boltzmann_parameter,
            n_troughs)

        trough_prior[~trough_thresholds] = 0

        # 5. For each threshold add probability to global minimum if no trough
        # is below threshold, else add probability to each trough below
        # threshold biased by prior.
        probs = trough_prior.dot(beta_probs)

        global_min = np.argmin(trough_heights)
        n_thresholds_below_min = np.count_nonzero(
            ~trough_thresholds[global_min, :])
        probs[global_min] += no_trough_prob * np.sum(
            beta_probs[:n_thresholds_below_min])

        yin_probs[trough_index, i] = probs

    yin_period, frame_index = np.nonzero(yin_probs)

    # Refine peak by parabolic interpolation.
    period_candidates = min_period + yin_period
    period_candidates = period_candidates + \
        parabolic_shifts[yin_period, frame_index]
    f0_candidates = penn.SAMPLE_RATE / period_candidates

    # Find pitch bin corresponding to each f0 candidate.
    bin_index = 12 * n_bins_per_semitone * np.log2(f0_candidates / penn.FMIN)
    bin_index = np.clip(
        np.round(bin_index),
        0,
        penn.PITCH_BINS - 1).astype(int)

    # Observation probabilities.
    observation_probs = np.zeros((penn.PITCH_BINS, frames.shape[1]))
    observation_probs[bin_index, frame_index] = \
        yin_probs[yin_period, frame_index]

    return observation_probs[np.newaxis]



import torch

import penn


###############################################################################
# Constants
###############################################################################


# Amount of bin downsampling
DOWNSAMPLE_RATE = penn.PITCH_BINS // 90


###############################################################################
# Plot dataset density vs true positive density
###############################################################################


def to_file(
    datasets,
    output_file,
    checkpoint=None,
    gpu=None):
    """Plot ground truth and true positive densities"""
    import matplotlib
    import matplotlib.pyplot as plt
    matplotlib.rcParams.update({'font.size': 20})
    figure, axis = plt.subplots()
    axis.set_axis_off()

    # Plot true data density
    x = torch.arange(0, penn.PITCH_BINS, DOWNSAMPLE_RATE)
    y_true, y_pred = histograms(datasets, checkpoint, gpu)
    y_true = y_true.reshape(-1, DOWNSAMPLE_RATE).sum(-1)
    axis.bar(
        x,
        y_true,
        width=DOWNSAMPLE_RATE,
        label=f'Data distribution')

    # Plot our guesses
    y_pred = y_pred.reshape(-1, DOWNSAMPLE_RATE).sum(-1)
    axis.bar(
        x,
        y_pred,
        width=DOWNSAMPLE_RATE,
        label='Inferred distribution')

    # Plot overlap
    overlap = torch.minimum(y_true, y_pred)
    axis.bar(
        x,
        overlap,
        color='gray',
        width=DOWNSAMPLE_RATE,
        label='Overlap')

    # Add legend
    axis.legend(frameon=False, prop={'size': 10})

    # Save plot
    figure.savefig(output_file, bbox_inches='tight', pad_inches=0, dpi=300)


def histograms(datasets, checkpoint=None, gpu=None):
    """Get histogram of true positives from datasets and model checkpoint"""
    device = torch.device('cpu' if gpu is None else f'cuda:{gpu}')

    # Initialize counts
    true_result = torch.zeros((penn.PITCH_BINS,))
    infer_result = torch.zeros((penn.PITCH_BINS,))

    # Setup loader
    loader = penn.data.loader(datasets, 'test', gpu)

    # Update counts
    for audio, bins, pitch, voiced, _ in loader:

        # Preprocess audio
        batch_size = \
            None if gpu is None else penn.EVALUATION_BATCH_SIZE
        iterator = penn.preprocess(
            audio[0],
            penn.SAMPLE_RATE,
            batch_size=batch_size,
            pad=True)
        for i, (frames, size) in enumerate(iterator):

            # Copy to device
            frames = frames.to(device)

            # Slice features and copy to GPU
            start = i * penn.EVALUATION_BATCH_SIZE
            end = start + size
            batch_bins = bins[:, start:end].to(device)
            batch_pitch = pitch[:, start:end].to(device)
            batch_voiced = voiced[:, start:end].to(device)

            # Infer
            batch_logits = penn.infer(frames, checkpoint).detach()

            # Get predicted bins
            batch_predicted, _, _ = penn.postprocess(batch_logits)

            # Get true positives
            true_all = batch_bins[batch_voiced]
            pred_all = batch_predicted[batch_voiced]

            # Update counts
            indices = torch.arange(
                penn.PITCH_BINS + 1,
                dtype=torch.float,
                device=device)
            true_result += torch.histogram(true_all.cpu().float(), indices.cpu())[0]
            infer_result += torch.histogram(pred_all.cpu().float(), indices.cpu())[0]

    return true_result, infer_result



import torch

import penn


###############################################################################
# Constants
###############################################################################


# Evaluation threshold for RPA and RCA
THRESHOLD = 50  # cents


###############################################################################
# Aggregate metric
###############################################################################


class Metrics:

    def __init__(self):
        self.accuracy = Accuracy()
        self.f1 = F1()
        self.loss = Loss()
        self.pitch_metrics = PitchMetrics()

    def __call__(self):
        return (
            self.accuracy() |
            self.f1() |
            self.loss() |
            self.pitch_metrics())

    def update(self, logits, bins, target, voiced):
        # Detach from graph
        logits = logits.detach()

        # Update loss
        self.loss.update(logits[:, :penn.PITCH_BINS], bins.T)

        # Decode bins, pitch, and periodicity
        with penn.time.timer('decode'):
            predicted, pitch, periodicity = penn.postprocess(logits)

        # Update bin accuracy
        self.accuracy.update(predicted[voiced], bins[voiced])

        # Update pitch metrics
        self.pitch_metrics.update(pitch, target, voiced)

        # Update periodicity metrics
        self.f1.update(periodicity, voiced)

    def reset(self):
        self.accuracy.reset()
        self.f1.reset()
        self.loss.reset()
        self.pitch_metrics.reset()


class PitchMetrics:

    def __init__(self):
        self.l1 = L1()
        self.rca = RCA()
        self.rmse = RMSE()
        self.rpa = RPA()

    def __call__(self):
        return self.l1() | self.rca() | self.rmse() | self.rpa()

    def update(self, pitch, target, voiced):
        # Mask unvoiced
        pitch, target = pitch[voiced], target[voiced]

        # Update metrics
        self.l1.update(pitch, target)
        self.rca.update(pitch, target)
        self.rmse.update(pitch, target)
        self.rpa.update(pitch, target)

    def reset(self):
        self.l1.reset()
        self.rca.reset()
        self.rmse.reset()
        self.rpa.reset()


###############################################################################
# Individual metrics
###############################################################################


class Accuracy:

    def __init__(self):
        self.reset()

    def __call__(self):
        return {'accuracy': (self.true_positives / self.count).item()}

    def update(self, predicted, target):
        self.true_positives += (predicted == target).sum()
        self.count += predicted.shape[-1]

    def reset(self):
        self.true_positives = 0
        self.count = 0


class F1:

    def __init__(self, thresholds=None):
        if thresholds is None:
            thresholds = sorted(list(set(
                [2 ** -i for i in range(1, 11)] +
                [.1 * i for i in range(10)])))
        self.thresholds = thresholds
        self.precision = [Precision() for _ in range(len(thresholds))]
        self.recall = [Recall() for _ in range(len(thresholds))]

    def __call__(self):
        result = {}
        iterator = zip(self.thresholds, self.precision, self.recall)
        for threshold, precision, recall in iterator:
            precision = precision()['precision']
            recall = recall()['recall']
            try:
                f1 = 2 * precision * recall / (precision + recall)
            except ZeroDivisionError:
                f1 = 0.
            result |= {
                f'f1-{threshold:.6f}': f1,
                f'precision-{threshold:.6f}': precision,
                f'recall-{threshold:.6f}': recall}
        return result

    def update(self, periodicity, voiced):
        iterator = zip(self.thresholds, self.precision, self.recall)
        for threshold, precision, recall in iterator:
            predicted = penn.voicing.threshold(periodicity, threshold)
            precision.update(predicted, voiced)
            recall.update(predicted, voiced)

    def reset(self):
        """Reset the F1 score"""
        for precision, recall in zip(self.precision, self.recall):
            precision.reset()
            recall.reset()


class L1:
    """L1 pitch distance in cents"""

    def __init__(self):
        self.reset()

    def __call__(self):
        return {'l1': (self.sum / self.count).item()}

    def update(self, predicted, target):
        self.sum += torch.abs(penn.cents(predicted, target)).sum()
        self.count += predicted.shape[-1]

    def reset(self):
        self.count = 0
        self.sum = 0.


class Loss():

    def __init__(self):
        self.reset()

    def __call__(self):
        return {'loss': (self.total / self.count).item()}

    def update(self, logits, bins):
        self.total += penn.train.loss(logits, bins)
        self.count += bins.shape[0]

    def reset(self):
        self.count = 0
        self.total = 0.


class Precision:

    def __init__(self):
        self.reset()

    def __call__(self):
        precision = (
            self.true_positives /
            (self.true_positives + self.false_positives)).item()
        return {'precision': precision}

    def update(self, predicted, voiced):
        self.true_positives += (predicted & voiced).sum()
        self.false_positives += (predicted & ~voiced).sum()

    def reset(self):
        self.true_positives = 0
        self.false_positives = 0


class Recall:

    def __init__(self):
        self.reset()

    def __call__(self):
        recall = (
            self.true_positives /
            (self.true_positives + self.false_negatives)).item()
        return {'recall': recall}

    def update(self, predicted, voiced):
        self.true_positives += (predicted & voiced).sum()
        self.false_negatives += (~predicted & voiced).sum()

    def reset(self):
        self.true_positives = 0
        self.false_negatives = 0


class RCA:
    """Raw chroma accuracy"""

    def __init__(self):
        self.reset()

    def __call__(self):
        return {'rca': (self.sum / self.count).item()}

    def update(self, predicted, target):
        # Compute pitch difference in cents
        difference = penn.cents(predicted, target)

        # Forgive octave errors
        difference[difference > (penn.OCTAVE - THRESHOLD)] -= penn.OCTAVE
        difference[difference < -(penn.OCTAVE - THRESHOLD)] += penn.OCTAVE

        # Count predictions that are within 50 cents of target
        self.sum += (torch.abs(difference) < THRESHOLD).sum()
        self.count += predicted.shape[-1]

    def reset(self):
        self.count = 0
        self.sum = 0


class RMSE:
    """Root mean square error of pitch distance in cents"""

    def __init__(self):
        self.reset()

    def __call__(self):
        return {'rmse': torch.sqrt(self.sum / self.count).item()}

    def update(self, predicted, target):
        self.sum += (penn.cents(predicted, target) ** 2).sum()
        self.count += predicted.shape[-1]

    def reset(self):
        self.count = 0
        self.sum = 0.


class RPA:
    """Raw prediction accuracy"""

    def __init__(self):
        self.reset()

    def __call__(self):
        return {'rpa': (self.sum / self.count).item()}

    def update(self, predicted, target):
        difference = penn.cents(predicted, target)
        self.sum += (torch.abs(difference) < THRESHOLD).sum()
        self.count += predicted.shape[-1]

    def reset(self):
        self.count = 0
        self.sum = 0



import json
import math
import tempfile
import time
from pathlib import Path

import numpy as np
import torch

import penn


###############################################################################
# Evaluate
###############################################################################


def datasets(
    datasets=penn.EVALUATION_DATASETS,
    checkpoint=penn.DEFAULT_CHECKPOINT,
    gpu=None):
    """Perform evaluation"""
    # Make output directory
    directory = penn.EVAL_DIR / penn.CONFIG
    directory.mkdir(exist_ok=True, parents=True)

    # Evaluate pitch estimation quality and save logits
    pitch_quality(directory, datasets, checkpoint, gpu)

    with tempfile.TemporaryDirectory() as directory:
        directory = Path(directory)

        # Get periodicity methods
        if penn.METHOD == 'dio':
            periodicity_fns = {}
        elif penn.METHOD == 'pyin':
            periodicity_fns = {'sum': penn.periodicity.sum}
        else:
            periodicity_fns = {
                'entropy': penn.periodicity.entropy,
                'max': penn.periodicity.max}

        # Evaluate periodicity
        periodicity_results = {}
        for key, val in periodicity_fns.items():
            periodicity_results[key] = periodicity_quality(
                directory,
                val,
                datasets,
                checkpoint=checkpoint,
                gpu=gpu)

        # Write periodicity results
        file = penn.EVAL_DIR / penn.CONFIG / 'periodicity.json'
        with open(file, 'w') as file:
            json.dump(periodicity_results, file, indent=4)

    # Perform benchmarking on CPU
    benchmark_results = {'cpu': benchmark(datasets, checkpoint)}

    # PYIN and DIO do not have GPU support
    if penn.METHOD not in ['dio', 'pyin']:
        benchmark_results ['gpu'] = benchmark(datasets, checkpoint, gpu)

    # Write benchmarking information
    with open(penn.EVAL_DIR / penn.CONFIG / 'time.json', 'w') as file:
        json.dump(benchmark_results, file, indent=4)


###############################################################################
# Individual evaluations
###############################################################################


def benchmark(
    datasets=penn.EVALUATION_DATASETS,
    checkpoint=penn.DEFAULT_CHECKPOINT,
    gpu=None):
    """Perform benchmarking"""
    # Get audio files
    dataset_stems = {
        dataset: penn.load.partition(dataset)['test'] for dataset in datasets}
    files = [
        penn.CACHE_DIR / dataset / f'{stem}.wav'
        for dataset, stems in dataset_stems.items()
        for stem in stems]

    # Setup temporary directory
    with tempfile.TemporaryDirectory() as directory:
        directory = Path(directory)

        # Create output directories
        for dataset in datasets:
            (directory / dataset).mkdir(exist_ok=True, parents=True)

        # Get output prefixes
        output_prefixes = [
            directory / file.parent.name / file.stem for file in files]

        # Start benchmarking
        penn.BENCHMARK = True
        penn.TIMER.reset()
        start_time = time.time()

        # Infer to temporary storage
        if penn.METHOD == 'penn':
            batch_size = \
                    None if gpu is None else penn.EVALUATION_BATCH_SIZE
            penn.from_files_to_files(
                files,
                output_prefixes,
                checkpoint=checkpoint,
                batch_size=batch_size,
                pad=True,
                gpu=gpu)

        elif penn.METHOD == 'torchcrepe':

            import torchcrepe

            # Get output file paths
            pitch_files = [
                file.parent / f'{file.stem}-pitch.pt'
                for file in output_prefixes]
            periodicity_files = [
                file.parent / f'{file.stem}-periodicity.pt'
                for file in output_prefixes]

            # Infer
            # Note - this does not perform the correct padding, but suffices
            #        for benchmarking purposes
            batch_size = \
                    None if gpu is None else penn.EVALUATION_BATCH_SIZE
            torchcrepe.predict_from_files_to_files(
                files,
                pitch_files,
                output_periodicity_files=periodicity_files,
                hop_length=penn.HOPSIZE,
                decoder=torchcrepe.decode.argmax,
                batch_size=batch_size,
                device='cpu' if gpu is None else f'cuda:{gpu}',
                pad=False)
        elif penn.METHOD == 'dio':
            penn.dsp.dio.from_files_to_files(files, output_prefixes)
        elif penn.METHOD == 'pyin':
            penn.dsp.pyin.from_files_to_files(files, output_prefixes)

        # Turn off benchmarking
        penn.BENCHMARK = False

        # Get benchmarking information
        benchmark = penn.TIMER()
        benchmark['total'] = time.time() - start_time

    # Get total number of samples and seconds in test data
    samples = sum([
        len(np.load(file.parent / f'{file.stem}-audio.npy', mmap_mode='r'))
        for file in files])
    seconds = penn.convert.samples_to_seconds(samples)

    # Format benchmarking results
    return {
        key: {
            'real-time-factor': value / seconds,
            'samples': samples,
            'samples-per-second': samples / value,
            'seconds': value
        } for key, value in benchmark.items()}


def periodicity_quality(
        directory,
        periodicity_fn,
        datasets=penn.EVALUATION_DATASETS,
        steps=8,
        checkpoint=penn.DEFAULT_CHECKPOINT,
        gpu=None):
    """Fine-grained periodicity estimation quality evaluation"""
    device = torch.device('cpu' if gpu is None else f'cuda:{gpu}')

    # Evaluate each dataset
    for dataset in datasets:

        # Create output directory
        (directory / dataset).mkdir(exist_ok=True, parents=True)

        # Setup dataset
        iterator = penn.iterator(
            penn.data.loader([dataset], 'valid', gpu, True),
            f'Evaluating {penn.CONFIG} periodicity quality on {dataset}')

        # Iterate over validation set
        for audio, _, _, voiced, stem in iterator:

            if penn.METHOD == 'penn':

                # Accumulate logits
                logits = []

                # Preprocess audio
                batch_size = \
                    None if gpu is None else penn.EVALUATION_BATCH_SIZE
                iterator = penn.preprocess(
                    audio[0],
                    penn.SAMPLE_RATE,
                    batch_size=batch_size,
                    pad=True)
                for frames, _ in iterator:

                    # Copy to device
                    frames = frames.to(device)

                    # Infer
                    batch_logits = penn.infer(frames, checkpoint).detach()

                    # Accumulate logits
                    logits.append(batch_logits)

                logits = torch.cat(logits)

            elif penn.METHOD == 'torchcrepe':

                import torchcrepe

                # Accumulate logits
                logits = []

                # Postprocessing breaks gradients, so just don't compute them
                with torch.no_grad():

                    # Preprocess audio
                    batch_size = \
                        None if gpu is None else penn.EVALUATION_BATCH_SIZE
                    pad = (penn.WINDOW_SIZE - penn.HOPSIZE) // 2
                    generator = torchcrepe.preprocess(
                        torch.nn.functional.pad(audio, (pad, pad))[0],
                        penn.SAMPLE_RATE,
                        penn.HOPSIZE,
                        batch_size,
                        device,
                        False)
                    for frames in generator:

                        # Infer independent probabilities for each pitch bin
                        batch_logits = torchcrepe.infer(
                            frames.to(device))[:, :, None]

                        # Accumulate logits
                        logits.append(batch_logits)
                    logits = torch.cat(logits)

            elif penn.METHOD == 'pyin':

                # Pad
                pad = (penn.WINDOW_SIZE - penn.HOPSIZE) // 2
                audio = torch.nn.functional.pad(audio, (pad, pad))

                # Infer
                logits = penn.dsp.pyin.infer(audio[0])

            # Save to temporary storage
            file = directory / dataset / f'{stem[0]}-logits.pt'
            torch.save(logits, file)

    # Default values
    best_threshold = .5
    best_value = 0.
    stepsize = .05

    # Setup metrics
    metrics = penn.evaluate.metrics.F1()

    step = 0
    while step < steps:

        for dataset in datasets:

            # Setup loader
            loader = penn.data.loader([dataset], 'valid', gpu, True)

            # Iterate over validation set
            for _, _, _, voiced, stem in loader:

                # Load logits
                logits = torch.load(directory / dataset / f'{stem[0]}-logits.pt')

                # Decode periodicity
                periodicity = periodicity_fn(logits.to(device)).T

                # Update metrics
                metrics.update(periodicity, voiced.to(device))

        # Get best performing threshold
        results = {
            key: val for key, val in metrics().items() if key.startswith('f1')
            and not math.isnan(val)}
        key = max(results, key=results.get)
        threshold = float(key[3:])
        value = results[key]
        if value > best_value:
            best_value = value
            best_threshold = threshold

        # Reinitialize metrics with new thresholds
        metrics = penn.evaluate.metrics.F1(
            [best_threshold - stepsize, best_threshold + stepsize])

        # Binary search for optimal threshold
        stepsize /= 2
        step += 1

    # Setup metrics with optimal threshold
    metrics = penn.evaluate.metrics.F1([best_threshold])

    # Setup test loader
    loader = penn.data.loader(datasets, 'test', gpu)

    # Iterate over test set
    for audio, _, _, voiced, stem in loader:

        if penn.METHOD == 'penn':

            # Accumulate logits
            logits = []

            # Preprocess audio
            batch_size = \
                None if gpu is None else penn.EVALUATION_BATCH_SIZE
            iterator = penn.preprocess(
                audio[0],
                penn.SAMPLE_RATE,
                batch_size=batch_size,
                pad=True)
            for frames, _ in iterator:

                # Copy to device
                frames = frames.to(device)

                # Infer
                batch_logits = penn.infer(frames, checkpoint).detach()

                # Accumulate logits
                logits.append(batch_logits)

            logits = torch.cat(logits)

        elif penn.METHOD == 'torchcrepe':

            import torchcrepe

            # Accumulate logits
            logits = []

            # Postprocessing breaks gradients, so just don't compute them
            with torch.no_grad():

                # Preprocess audio
                batch_size = \
                    None if gpu is None else penn.EVALUATION_BATCH_SIZE
                pad = (penn.WINDOW_SIZE - penn.HOPSIZE) // 2
                generator = torchcrepe.preprocess(
                    torch.nn.functional.pad(audio, (pad, pad))[0],
                    penn.SAMPLE_RATE,
                    penn.HOPSIZE,
                    batch_size,
                    device,
                    False)
                for frames in generator:

                    # Infer independent probabilities for each pitch bin
                    batch_logits = torchcrepe.infer(
                        frames.to(device))[:, :, None]

                    # Accumulate logits
                    logits.append(batch_logits)
                logits = torch.cat(logits)

        elif penn.METHOD == 'pyin':

            # Pad
            pad = (penn.WINDOW_SIZE - penn.HOPSIZE) // 2
            audio = torch.nn.functional.pad(audio, (pad, pad))

            # Infer
            logits = penn.dsp.pyin.infer(audio[0]).to(device)

        # Decode periodicity
        periodicity = periodicity_fn(logits).T

        # Update metrics
        metrics.update(periodicity, voiced.to(device))

    # Get F1 score on test set
    score = metrics()[f'f1-{best_threshold:.6f}']

    return {'threshold': best_threshold, 'f1': score}


def pitch_quality(
    directory,
    datasets=penn.EVALUATION_DATASETS,
    checkpoint=penn.DEFAULT_CHECKPOINT,
    gpu=None):
    """Evaluate pitch estimation quality"""
    device = torch.device('cpu' if gpu is None else f'cuda:{gpu}')

    # Containers for results
    overall, granular = {}, {}

    # Get metric class
    metric_fn = (
        penn.evaluate.PitchMetrics if penn.METHOD == 'dio' else
        penn.evaluate.Metrics)

    # Per-file metrics
    file_metrics = metric_fn()

    # Per-dataset metrics
    dataset_metrics = metric_fn()

    # Aggregate metrics over all datasets
    aggregate_metrics = metric_fn()

    # Evaluate each dataset
    for dataset in datasets:

        # Reset dataset metrics
        dataset_metrics.reset()

        # Setup test dataset
        iterator = penn.iterator(
            penn.data.loader([dataset], 'test'),
            f'Evaluating {penn.CONFIG} pitch quality on {dataset}')

        # Iterate over test set
        for audio, bins, pitch, voiced, stem in iterator:

            # Reset file metrics
            file_metrics.reset()

            if penn.METHOD == 'penn':

                # Accumulate logits
                logits = []

                # Preprocess audio
                batch_size = \
                    None if gpu is None else penn.EVALUATION_BATCH_SIZE
                iterator = penn.preprocess(
                    audio[0],
                    penn.SAMPLE_RATE,
                    batch_size=batch_size,
                    pad=True)
                for i, (frames, size) in enumerate(iterator):

                    # Copy to device
                    frames = frames.to(device)

                    # Slice features and copy to GPU
                    start = i * penn.EVALUATION_BATCH_SIZE
                    end = start + size
                    batch_bins = bins[:, start:end].to(device)
                    batch_pitch = pitch[:, start:end].to(device)
                    batch_voiced = voiced[:, start:end].to(device)

                    # Infer
                    batch_logits = penn.infer(frames, checkpoint).detach()

                    # Update metrics
                    args = (
                        batch_logits,
                        batch_bins,
                        batch_pitch,
                        batch_voiced)
                    file_metrics.update(*args)
                    dataset_metrics.update(*args)
                    aggregate_metrics.update(*args)

                    # Accumulate logits
                    logits.append(batch_logits)
                logits = torch.cat(logits)

            elif penn.METHOD == 'torchcrepe':

                import torchcrepe

                # Accumulate logits
                logits = []

                # Postprocessing breaks gradients, so just don't compute them
                with torch.no_grad():

                    # Preprocess audio
                    batch_size = \
                        None if gpu is None else penn.EVALUATION_BATCH_SIZE
                    pad = (penn.WINDOW_SIZE - penn.HOPSIZE) // 2
                    generator = torchcrepe.preprocess(
                        torch.nn.functional.pad(audio, (pad, pad))[0],
                        penn.SAMPLE_RATE,
                        penn.HOPSIZE,
                        batch_size,
                        device,
                        False)
                    for i, frames in enumerate(generator):

                        # Infer independent probabilities for each pitch bin
                        batch_logits = torchcrepe.infer(frames.to(device))[:, :, None]

                        # Slice features and copy to GPU
                        start = i * penn.EVALUATION_BATCH_SIZE
                        end = start + frames.shape[0]
                        batch_bins = bins[:, start:end].to(device)
                        batch_pitch = pitch[:, start:end].to(device)
                        batch_voiced = voiced[:, start:end].to(device)

                        # Update metrics
                        args = (
                            batch_logits,
                            batch_bins,
                            batch_pitch,
                            batch_voiced)
                        file_metrics.update(*args)
                        dataset_metrics.update(*args)
                        aggregate_metrics.update(*args)

                        # Accumulate logits
                        logits.append(batch_logits)
                    logits = torch.cat(logits)

            elif penn.METHOD == 'dio':

                # Pad
                pad = (penn.WINDOW_SIZE - penn.HOPSIZE) // 2
                audio = torch.nn.functional.pad(audio, (pad, pad))

                # Infer
                predicted = penn.dsp.dio.from_audio(audio[0])

                # Update metrics
                args = predicted, pitch, voiced
                file_metrics.update(*args)
                dataset_metrics.update(*args)
                aggregate_metrics.update(*args)

            elif penn.METHOD == 'pyin':

                # Pad
                pad = (penn.WINDOW_SIZE - penn.HOPSIZE) // 2
                audio = torch.nn.functional.pad(audio, (pad, pad))

                # Infer
                logits = penn.dsp.pyin.infer(audio[0])

                # Update metrics
                args = logits, bins, pitch, voiced
                file_metrics.update(*args)
                dataset_metrics.update(*args)
                aggregate_metrics.update(*args)

            # Copy results
            granular[f'{dataset}/{stem[0]}'] = file_metrics()
        overall[dataset] = dataset_metrics()
    overall['aggregate'] = aggregate_metrics()

    # Write to json files
    directory = penn.EVAL_DIR / penn.CONFIG
    with open(directory / 'overall.json', 'w') as file:
        json.dump(overall, file, indent=4)
    with open(directory / 'granular.json', 'w') as file:
        json.dump(granular, file, indent=4)



import functools
import multiprocessing as mp

import numpy as np
import torch

import penn


###############################################################################
# DIO (from pyworld)
###############################################################################


def from_audio(
    audio,
    sample_rate=penn.SAMPLE_RATE,
    hopsize=penn.HOPSIZE_SECONDS,
    fmin=penn.FMIN,
    fmax=penn.FMAX):
    """Estimate pitch and periodicity with dio"""
    with penn.time.timer('infer'):

        import pyworld

        # Convert to numpy
        audio = audio.numpy().squeeze().astype(np.float)

        # Get pitch
        pitch, times  = pyworld.dio(
            audio[penn.WINDOW_SIZE // 2:-penn.WINDOW_SIZE // 2],
            sample_rate,
            fmin,
            fmax,
            frame_period=1000 * hopsize)

        # Refine pitch
        pitch = pyworld.stonemask(
            audio,
            pitch,
            times,
            sample_rate)

        # Interpolate unvoiced tokens
        pitch, _ = penn.data.preprocess.interpolate_unvoiced(pitch)

        # Convert to torch
        return torch.from_numpy(pitch)[None]


def from_file(
        file,
        hopsize=penn.HOPSIZE_SECONDS,
        fmin=penn.FMIN,
        fmax=penn.FMAX):
    """Estimate pitch and periodicity with dio from audio on disk"""
    # Load
    with penn.time.timer('load'):
        audio = penn.load.audio(file)

    # Infer
    return from_audio(audio, penn.SAMPLE_RATE, hopsize, fmin, fmax)


def from_file_to_file(
        file,
        output_prefix=None,
        hopsize=penn.HOPSIZE_SECONDS,
        fmin=penn.FMIN,
        fmax=penn.FMAX):
    """Estimate pitch and periodicity with dio and save to disk"""
    # Infer
    results = from_file(file, hopsize, fmin, fmax)

    # Save to disk
    with penn.time.timer('save'):

        # Maybe use same filename with new extension
        if output_prefix is None:
            output_prefix = file.parent / file.stem

        # Save pitch
        torch.save(results[0], f'{output_prefix}-pitch.pt')

        # Maybe save periodicity
        if len(results) > 1:
            torch.save(results[1], f'{output_prefix}-periodicity.pt')


def from_files_to_files(
        files,
        output_prefixes=None,
        hopsize=penn.HOPSIZE_SECONDS,
        fmin=penn.FMIN,
        fmax=penn.FMAX):
    """Estimate pitch and periodicity with dio and save to disk"""
    pitch_fn = functools.partial(
        from_file_to_file,
        hopsize=hopsize,
        fmin=fmin,
        fmax=fmax)
    iterator = zip(files, output_prefixes)

    # Turn off multiprocessing for benchmarking
    if penn.BENCHMARK:
        iterator = penn.iterator(
            iterator,
            f'{penn.CONFIG}',
            total=len(files))
        for item in iterator:
            pitch_fn(*item)
    else:
        with mp.get_context('spawn').Pool() as pool:
            pool.starmap(pitch_fn, iterator)



import json
import random

import penn


###############################################################################
# Dataset-specific
###############################################################################


def datasets(datasets):
    """Partition datasets"""
    for name in datasets:
        dataset(name)


def dataset(name):
    """Partition dataset"""
    # Get dataset stems
    stems = sorted([
        file.stem[:-6] for file in
        (penn.CACHE_DIR / name).glob('*-audio.npy')])
    random.seed(penn.RANDOM_SEED)
    random.shuffle(stems)

    # Get split points
    left, right = int(.70 * len(stems)), int(.85 * len(stems))

    # Perform partition
    partition = {
        'train': sorted(stems[:left]),
        'valid': sorted(stems[left:right]),
        'test': sorted(stems[right:])}

    # Write partition file
    with open(penn.PARTITION_DIR / f'{name}.json', 'w') as file:
        json.dump(partition, file, indent=4)



import torch

import penn


def loader(datasets, partition, gpu=None, hparam_search=False):
    """Retrieve a data loader"""
    # Create dataset
    dataset = penn.data.Dataset(datasets, partition, hparam_search)

    # Create sampler
    sampler = penn.data.sampler(dataset, partition)

    # Get batch size
    if partition == 'train':

        # Maybe split batch over GPUs
        if torch.distributed.is_initialized():
            batch_size = penn.BATCH_SIZE // torch.distributed.get_world_size()
        else:
            batch_size = penn.BATCH_SIZE

    elif partition == 'test' or (partition == 'valid' and hparam_search):
        batch_size = 1
    elif partition == 'valid':
        batch_size = penn.BATCH_SIZE
    else:
        raise ValueError(f'Partition {partition} is not defined')

    # Create data loader
    return torch.utils.data.DataLoader(
        dataset=dataset,
        batch_size=batch_size,
        num_workers=penn.NUM_WORKERS,
        pin_memory=gpu is not None,
        sampler=sampler)



import argparse
import shutil
from pathlib import Path

import penn


###############################################################################
# Entry point
###############################################################################


def main(config, datasets, gpus=None):
    # Create output directory
    directory = penn.RUNS_DIR / config.stem
    directory.mkdir(parents=True, exist_ok=True)

    # Save configuration
    shutil.copyfile(config, directory / config.name)

    # Train
    checkpoint = penn.train.run(
        datasets,
        directory,
        directory,
        directory,
        gpus)

    # Evaluate
    penn.evaluate.datasets(checkpoint=checkpoint, gpu=gpus[0])


def parse_args():
    """Parse command-line arguments"""
    parser = argparse.ArgumentParser(description='Train a model')
    parser.add_argument(
        '--config',
        type=Path,
        required=True,
        help='The configuration file')
    parser.add_argument(
        '--datasets',
        nargs='+',
        default=penn.DATASETS,
        help='The datasets to train on')
    parser.add_argument(
        '--gpus',
        type=int,
        nargs='+',
        help='The gpus to run training on')
    return parser.parse_args()


main(**vars(parse_args()))



import contextlib
import functools
import os

import torch

import penn


###############################################################################
# Training interface
###############################################################################


def run(
    datasets,
    checkpoint_directory,
    output_directory,
    log_directory,
    gpus=None):
    """Run model training"""
    # Distributed data parallelism
    if gpus and len(gpus) > 1:
        args = (
            datasets,
            checkpoint_directory,
            output_directory,
            log_directory,
            gpus)
        torch.multiprocessing.spawn(
            train_ddp,
            args=args,
            nprocs=len(gpus),
            join=True)

    else:

        # Single GPU or CPU training
        train(
            datasets,
            checkpoint_directory,
            output_directory,
            log_directory,
            None if gpus is None else gpus[0])

    # Return path to generator checkpoint
    return penn.checkpoint.latest_path(output_directory)


###############################################################################
# Training
###############################################################################


def train(
    datasets,
    checkpoint_directory,
    output_directory,
    log_directory,
    gpu=None):
    """Train a model"""
    # Get DDP rank
    if torch.distributed.is_initialized():
        rank = torch.distributed.get_rank()
    else:
        rank = None

    # Get torch device
    device = torch.device('cpu' if gpu is None else f'cuda:{gpu}')

    #######################
    # Create data loaders #
    #######################

    torch.manual_seed(penn.RANDOM_SEED)
    train_loader = penn.data.loader(datasets, 'train', gpu)
    valid_loader = penn.data.loader(datasets, 'valid', gpu)

    ################
    # Create model #
    ################

    model = penn.model.Model().to(device)

    ####################
    # Create optimizer #
    ####################

    optimizer = torch.optim.Adam(model.parameters(), lr=penn.LEARNING_RATE)

    ##############################
    # Maybe load from checkpoint #
    ##############################

    path = penn.checkpoint.latest_path(checkpoint_directory, '*.pt')

    if path is not None:

        # Load model
        model, optimizer, step = penn.checkpoint.load(path, model, optimizer)

    else:

        # Train from scratch
        step = 0

    ##################################################
    # Maybe setup distributed data parallelism (DDP) #
    ##################################################

    if rank is not None:
        model = torch.nn.parallel.DistributedDataParallel(
            model,
            device_ids=[rank])

    ##############################
    # Maybe setup early stopping #
    ##############################

    if penn.EARLY_STOPPING:
        counter = penn.EARLY_STOPPING_STEPS
        best_accuracy = 0.
        stop = False

    #########
    # Train #
    #########

    # Automatic mixed precision (amp) gradient scaler
    scaler = torch.cuda.amp.GradScaler()

    # Get total number of steps
    steps = penn.STEPS

    # Setup progress bar
    if not rank:
        progress = penn.iterator(
            range(step, steps),
            f'Training {penn.CONFIG}',
            step,
            steps)
    while step < steps and (not penn.EARLY_STOPPING or not stop):

        # Seed sampler
        epoch = step // len(train_loader.dataset)
        train_loader.sampler.set_epoch(epoch)

        for batch in train_loader:

            # Unpack batch
            audio, bins, *_ = batch

            with torch.autocast(device.type):

                # Forward pass
                logits = model(audio.to(device))

                # Compute losses
                losses = loss(logits, bins.to(device))

            ##################
            # Optimize model #
            ##################

            optimizer.zero_grad()

            # Backward pass
            scaler.scale(losses).backward()

            # Update weights
            scaler.step(optimizer)

            # Update gradient scaler
            scaler.update()

            ##############
            # Evaluation #
            ##############

            if not rank:

                # Save checkpoint
                if step and step % penn.CHECKPOINT_INTERVAL == 0:
                    penn.checkpoint.save(
                        model,
                        optimizer,
                        step,
                        output_directory / f'{step:08d}.pt')

                # Evaluate
                if step % penn.LOG_INTERVAL == 0:
                    evaluate_fn = functools.partial(
                        evaluate,
                        log_directory,
                        step,
                        model,
                        gpu)
                    evaluate_fn('train', train_loader)
                    valid_accuracy = evaluate_fn('valid', valid_loader)

                    # Maybe stop training
                    if penn.EARLY_STOPPING:
                        counter -= 1

                        # Update best validation loss
                        if valid_accuracy > best_accuracy:
                            best_accuracy = valid_accuracy
                            counter = penn.EARLY_STOPPING_STEPS

                        # Stop training
                        elif counter == 0:
                            stop = True

            # Update training step count
            if step >= steps or (penn.EARLY_STOPPING and stop):
                break
            step += 1

            # Update progress bar
            if not rank:
                progress.update()

    # Close progress bar
    if not rank:
        progress.close()

    # Save final model
    penn.checkpoint.save(
        model,
        optimizer,
        step,
        output_directory / f'{step:08d}.pt')


###############################################################################
# Evaluation
###############################################################################


def evaluate(directory, step, model, gpu, condition, loader):
    """Perform model evaluation"""
    device = torch.device('cpu' if gpu is None else f'cuda:{gpu}')

    # Setup evaluation metrics
    metrics = penn.evaluate.Metrics()

    # Prepare model for inference
    with penn.inference_context(model):

        # Unpack batch
        for i, (audio, bins, pitch, voiced, *_) in enumerate(loader):

            # Forward pass
            logits = model(audio.to(device))

            # Update metrics
            metrics.update(
                logits,
                bins.to(device).T,
                pitch.to(device).T,
                voiced.to(device).T)

            # Stop when we exceed some number of batches
            if i + 1 == penn.LOG_STEPS:
                break

    # Format results
    scalars = {
        f'{key}/{condition}': value for key, value in metrics().items()}

    # Write to tensorboard
    penn.write.scalars(directory, step, scalars)

    return scalars[f'accuracy/{condition}']


###############################################################################
# Distributed data parallelism
###############################################################################


def train_ddp(
    rank,
    dataset,
    checkpoint_directory,
    output_directory,
    log_directory,
    gpus):
    """Train with distributed data parallelism"""
    with ddp_context(rank, len(gpus)):
        train(
            dataset,
            checkpoint_directory,
            output_directory,
            log_directory,
            gpus[rank])


@contextlib.contextmanager
def ddp_context(rank, world_size):
    """Context manager for distributed data parallelism"""
    # Setup ddp
    os.environ['MASTER_ADDR']='localhost'
    os.environ['MASTER_PORT']='12355'
    torch.distributed.init_process_group(
        'nccl',
        init_method='env://',
        world_size=world_size,
        rank=rank)

    try:

        # Execute user code
        yield

    finally:

        # Close ddp
        torch.distributed.destroy_process_group()


###############################################################################
# Loss function
###############################################################################


def loss(logits, bins):
    """Compute loss function"""
    # Reshape inputs
    logits = logits.permute(0, 2, 1).reshape(-1, penn.PITCH_BINS)
    bins = bins.flatten()

    # Maybe blur target
    if penn.GAUSSIAN_BLUR:

        # Cache cents values to evaluate distributions at
        if not hasattr(loss, 'cents'):
            loss.cents = penn.convert.bins_to_cents(
                torch.arange(penn.PITCH_BINS))[:, None]

        # Ensure values are on correct device (no-op if devices are the same)
        loss.cents = loss.cents.to(bins.device)

        # Create normal distributions
        distributions = torch.distributions.Normal(
            penn.convert.bins_to_cents(bins),
            25)

        # Sample normal distributions
        bins = torch.exp(distributions.log_prob(loss.cents)).permute(1, 0)

        # Normalize
        bins = bins / (bins.max(dim=1, keepdims=True).values + 1e-8)

    else:

        # One-hot encoding
        bins = torch.nn.functional.one_hot(bins, penn.PITCH_BINS).float()

    if penn.LOSS == 'binary_cross_entropy':

        # Compute binary cross-entropy loss
        return torch.nn.functional.binary_cross_entropy_with_logits(
            logits,
            bins)

    elif penn.LOSS == 'categorical_cross_entropy':

        # Compute categorical cross-entropy loss
        return torch.nn.functional.cross_entropy(logits, bins)

    else:

        raise ValueError(f'Loss {penn.LOSS} is not implemented')



import numpy as np
import torch

import penn


###############################################################################
# Decode pitch contour from logits of pitch posteriorgram
###############################################################################


def argmax(logits):
    """Decode pitch using argmax"""
    # Get pitch bins
    bins = logits.argmax(dim=1)

    # Convert to hz
    pitch = penn.convert.bins_to_frequency(bins)

    return bins, pitch


def viterbi(logits):
    """Decode pitch using viterbi decoding (from librosa)"""
    import librosa

    # Normalize and convert to numpy
    if penn.METHOD == 'pyin':
        periodicity = penn.periodicity.sum(logits).T
        unvoiced = (
            (1 - periodicity) / penn.PITCH_BINS).repeat(penn.PITCH_BINS, 1)
        distributions = torch.cat(
            (torch.exp(logits.permute(2, 1, 0)), unvoiced[None]),
            dim=1).numpy()
    else:

        # Viterbi REQUIRES a categorical distribution, even if the loss was BCE
        distributions = torch.nn.functional.softmax(logits, dim=1)
        distributions = distributions.permute(2, 1, 0).numpy()

    # Cache viterbi probabilities
    if not hasattr(viterbi, 'transition'):
        # Get number of bins per frame
        bins_per_octave = penn.OCTAVE / penn.CENTS_PER_BIN
        max_octaves_per_frame = \
            penn.MAX_OCTAVES_PER_SECOND * penn.HOPSIZE / penn.SAMPLE_RATE
        max_bins_per_frame = max_octaves_per_frame * bins_per_octave+ 1

        # Construct the within voicing transition probabilities
        viterbi.transition = librosa.sequence.transition_local(
            penn.PITCH_BINS,
            max_bins_per_frame,
            window='triangle',
            wrap=False)

        if penn.METHOD == 'pyin':

            # Add unvoiced probabilities
            viterbi.transition = np.kron(
                librosa.sequence.transition_loop(2, .99),
                viterbi.transition)

            # Uniform initial probabilities
            viterbi.initial = np.zeros(2 * penn.PITCH_BINS)
            viterbi.initial[penn.PITCH_BINS:] = 1 / penn.PITCH_BINS

        else:

            # Uniform initial probabilities
            viterbi.initial = np.full(penn.PITCH_BINS, 1 / penn.PITCH_BINS)

    # Viterbi decoding
    bins = librosa.sequence.viterbi(
        distributions,
        viterbi.transition,
        p_init=viterbi.initial)
    bins = torch.from_numpy(bins.astype(np.int32))

    # Convert to frequency in Hz
    if penn.DECODER.endswith('normal'):

        # Decode using an assumption of normality around to the viterbi path
        pitch = locally_normal_from_bins(bins, logits)

    else:

        # Argmax decoding
        pitch = penn.convert.bins_to_frequency(bins)

    if penn.METHOD == 'pyin':

        # Linearly interpolate unvoiced regions
        pitch[bins >= penn.PITCH_BINS] = 0
        pitch = penn.data.preprocess.interpolate_unvoiced(pitch.numpy())[0]
        pitch = torch.from_numpy(pitch)

    return bins.T, pitch.T


def locally_normal(logits, window=penn.LOCAL_PITCH_WINDOW_SIZE):
    """Decode pitch using a normal assumption around the argmax"""
    # Get center bins
    bins = logits.argmax(dim=1)

    return bins, locally_normal_from_bins(bins, logits, window)


###############################################################################
# Utilities
###############################################################################


def expected_value(logits, cents):
    """Expected value computation from logits"""
    # Get local distributions
    if penn.LOSS == 'categorical_cross_entropy':
        distributions = torch.nn.functional.softmax(logits, dim=1)
    elif penn.LOSS == 'binary_cross_entropy':
        distributions = torch.sigmoid(logits)
    else:
        raise ValueError(f'Loss {penn.LOSS} is not defined')

    # Pitch is expected value in cents
    pitch = (distributions * cents).sum(dim=1, keepdims=True)

    # BCE requires normalization
    if penn.LOSS == 'binary_cross_entropy':
        pitch = pitch / distributions.sum(dim=1)

    # Convert to hz
    return penn.convert.cents_to_frequency(pitch)


def locally_normal_from_bins(bins, logits, window=penn.LOCAL_PITCH_WINDOW_SIZE):
    """Decode pitch using normal assumption around argmax from bin indices"""
    # Pad
    padded = torch.nn.functional.pad(
        logits.squeeze(2),
        (window // 2, window // 2),
        value=-float('inf'))

    # Get indices
    indices = \
        bins.repeat(1, window) + torch.arange(window, device=bins.device)[None]

    # Get values in cents
    cents = penn.convert.bins_to_cents(torch.clip(indices - window // 2, 0))

    # Decode using local expected value
    return expected_value(torch.gather(padded, 1, indices), cents)



import argparse

import penn


###############################################################################
# Download datasets
###############################################################################


def parse_args():
    """Parse command-line arguments"""
    parser = argparse.ArgumentParser(description='Download datasets')
    parser.add_argument(
        '--datasets',
        nargs='+',
        default=penn.DATASETS,
        help='The datasets to download')
    return parser.parse_args()


penn.data.download.datasets(**vars(parse_args()))



import shutil
import ssl
import tarfile
import urllib
import zipfile

import penn


###############################################################################
# Download datasets
###############################################################################


def datasets(datasets):
    """Download datasets"""
    if 'mdb' in datasets:
        mdb()

    if 'ptdb' in datasets:
        ptdb()


###############################################################################
# Individual datasets
###############################################################################


def mdb():
    """Download mdb dataset"""
    # Download
    url = 'https://zenodo.org/record/1481172/files/MDB-stem-synth.tar.gz'
    file = penn.SOURCE_DIR / 'mdb.tar.gz'
    download_file(url, file)

    with penn.chdir(penn.DATA_DIR):

        # Unzip
        with tarfile.open(file, 'r:gz') as tfile:
            tfile.extractall()

            # Delete previous directory
            shutil.rmtree('mdb', ignore_errors=True)

            # Rename directory
            shutil.move('MDB-stem-synth', 'mdb')


def ptdb():
    """Download ptdb dataset"""
    # Download
    url = 'https://www2.spsc.tugraz.at/databases/PTDB-TUG/SPEECH_DATA_ZIPPED.zip'
    file = penn.SOURCE_DIR / 'ptdb.zip'
    download_file(url, file)

    with penn.chdir(penn.DATA_DIR):

        # Unzip
        with zipfile.ZipFile(file, 'r') as zfile:
            zfile.extractall('ptdb')


###############################################################################
# Utilities
###############################################################################


def download_file(url, file):
    """Download file from url"""
    with urllib.request.urlopen(url, context=ssl.SSLContext()) as response, \
         open(file, 'wb') as output:
        shutil.copyfileobj(response, output)



import torch

import penn


###############################################################################
# Create figure
###############################################################################


def from_audio(
    audio,
    sample_rate,
    checkpoint=penn.DEFAULT_CHECKPOINT,
    gpu=None):
    """Plot logits with pitch overlay"""
    import matplotlib
    import matplotlib.pyplot as plt

    logits = []

    # Change font size
    matplotlib.rcParams.update({'font.size': 16})

    # Preprocess audio
    iterator = penn.preprocess(
        audio,
        sample_rate,
        batch_size=penn.EVALUATION_BATCH_SIZE,
        pad=True)
    for frames, _ in iterator:

        # Copy to device
        frames = frames.to('cpu' if gpu is None else f'cuda:{gpu}')

        # Infer
        logits.append(penn.infer(frames, checkpoint=checkpoint).detach())

    # Concatenate results
    logits = torch.cat(logits)

    # Convert to distribution
    # NOTE - We use softmax even if the loss is BCE for more comparable
    #        visualization. Otherwise, the variance of models trained with
    #        BCE looks erroneously lower.
    distributions = torch.nn.functional.softmax(logits, dim=1)

    # Take the log again for display
    distributions = torch.log(distributions)
    distributions[torch.isinf(distributions)] = \
        distributions[~torch.isinf(distributions)].min()

    # Prepare for plotting
    distributions = distributions.cpu().squeeze(2).T

    # Setup figure
    figure, axis = plt.subplots(figsize=(18, 2))

    # Make pretty
    axis.spines['top'].set_visible(False)
    axis.spines['right'].set_visible(False)
    axis.spines['bottom'].set_visible(False)
    axis.spines['left'].set_visible(False)
    xticks = torch.arange(0, len(logits), int(penn.SAMPLE_RATE / penn.HOPSIZE))
    xlabels = xticks // 100
    axis.get_xaxis().set_ticks(xticks.tolist(), xlabels.tolist())
    yticks = torch.linspace(0, penn.PITCH_BINS - 1, 5)
    ylabels = penn.convert.bins_to_frequency(yticks)
    ylabels = ylabels.round().int().tolist()
    axis.get_yaxis().set_ticks(yticks, ylabels)
    axis.set_xlabel('Time (seconds)')
    axis.set_ylabel('Frequency (Hz)')

    # Plot pitch posteriorgram
    axis.imshow(distributions, aspect='auto', origin='lower')

    return figure


def from_file(audio_file, checkpoint=penn.DEFAULT_CHECKPOINT, gpu=None):
    """Plot logits and optional pitch"""
    # Load audio
    audio = penn.load.audio(audio_file)

    # Plot
    return from_audio(audio, penn.SAMPLE_RATE, checkpoint, gpu)


def from_file_to_file(
    audio_file,
    output_file,
    checkpoint=penn.DEFAULT_CHECKPOINT,
    gpu=None):
    """Plot pitch and periodicity and save to disk"""
    # Plot
    figure = from_file(audio_file, checkpoint, gpu)

    # Save to disk
    figure.savefig(output_file, bbox_inches='tight', pad_inches=0, dpi=900)



# Evaluation
# - interpolate unvoiced


###############################################################################
# Configuration
###############################################################################


# Default configuration parameters to be modified
from .config import defaults

# Modify configuration
import yapecs
yapecs.configure('penn', defaults)

# Import configuration parameters
from .config.defaults import *
from . import time
from .config.static import *


###############################################################################
# Module imports
###############################################################################


from .core import *
from .model import Model
from . import checkpoint
from . import convert
from . import data
from . import decode
from . import dsp
from . import evaluate
from . import load
from . import partition
from . import periodicity
from . import plot
from . import train
from . import voicing
from . import write



import torch

import penn


class Fcnf0(torch.nn.Sequential):

    def __init__(self):
        layers = (penn.model.Normalize(),) if penn.NORMALIZE_INPUT else ()
        layers += (
            Block(1, 256, 481, (2, 2)),
            Block(256, 32, 225, (2, 2)),
            Block(32, 32, 97, (2, 2)),
            Block(32, 128, 66),
            Block(128, 256, 35),
            Block(256, 512, 4),
            torch.nn.Conv1d(512, penn.PITCH_BINS, 4))
        super().__init__(*layers)

    def forward(self, frames):
        # shape=(batch, 1, penn.WINDOW_SIZE) =>
        # shape=(batch, penn.PITCH_BINS, penn.NUM_TRAINING_FRAMES)
        return super().forward(frames[:, :, 16:-15])


class Block(torch.nn.Sequential):

    def __init__(
        self,
        in_channels,
        out_channels,
        length=1,
        pooling=None,
        kernel_size=32):
        layers = (
            torch.nn.Conv1d(in_channels, out_channels, kernel_size),
            torch.nn.ReLU())

        # Maybe add pooling
        if pooling is not None:
            layers += (torch.nn.MaxPool1d(*pooling),)

        # Maybe add normalization
        if penn.NORMALIZATION == 'batch':
            layers += (torch.nn.BatchNorm1d(out_channels, momentum=.01),)
        elif penn.NORMALIZATION == 'instance':
            layers += (torch.nn.InstanceNorm1d(out_channels),)
        elif penn.NORMALIZATION == 'layer':
            layers += (torch.nn.LayerNorm((out_channels, length)),)
        else:
            raise ValueError(
                f'Normalization method {penn.NORMALIZATION} is not defined')

        # Maybe add dropout
        if penn.DROPOUT is not None:
            layers += (torch.nn.Dropout(penn.DROPOUT),)

        super().__init__(*layers)



from torch.utils.tensorboard import SummaryWriter


###############################################################################
# Tensorboard logging
###############################################################################


def audio(directory, step, audio, sample_rate):
    """Write audio to Tensorboard"""
    for name, waveform in audio.items():
        writer(directory).add_audio(
            name,
            waveform,
            step,
            sample_rate)


def figures(directory, step, figures):
    """Write figures to Tensorboard"""
    for name, figure in figures.items():
        writer(directory).add_figure(name, figure, step)


def images(directory, step, images):
    """Write images to Tensorboard"""
    for name, image in images.items():
        writer(directory).add_image(name, image, step, dataformats='HWC')


def scalars(directory, step, scalars):
    """Write scalars to Tensorboard"""
    for name, scalar in scalars.items():
        writer(directory).add_scalar(name, scalar, step)


###############################################################################
# Utilities
###############################################################################


def writer(directory):
    """Get the writer object"""
    if not hasattr(writer, 'writer') or writer.directory != directory:
        writer.writer = SummaryWriter(log_dir=directory)
        writer.directory = directory
    return writer.writer



import json

import torchaudio

import penn


def audio(file):
    """Load audio from disk"""
    audio, sample_rate = torchaudio.load(file)

    # Maybe resample
    return penn.resample(audio, sample_rate)


def partition(dataset):
    """Load partitions for dataset"""
    with open(penn.PARTITION_DIR / f'{dataset}.json') as file:
        return json.load(file)



import torch

import penn


###############################################################################
# Crepe
###############################################################################


class Crepe(torch.nn.Sequential):

    def __init__(self):
        super().__init__()
        in_channels = [1, 1024, 128, 128, 128, 256]
        out_channels = [1024, 128, 128, 128, 256, 512]
        kernels = [512] + 5 * [64]
        strides = [4] + 5 * [1]
        padding = [(254, 254)] + 5 * [(31, 32)]
        lengths = [256, 128, 64, 32, 16, 8]
        super().__init__(*(
            ([penn.model.Normalize()] if penn.NORMALIZE_INPUT else []) +
            [
                Block(i, o, k, s, p, l) for i, o, k, s, p, l in
                zip(
                    in_channels,
                    out_channels,
                    kernels,
                    strides,
                    padding,
                    lengths
                )
            ] +
            [
                penn.model.Flatten(),
                torch.nn.Linear(
                    in_features=2048,
                    out_features=penn.PITCH_BINS)
            ]
        ))

    def forward(self, frames):
        # shape=(batch, 1, penn.WINDOW_SIZE) =>
        # shape=(batch, penn.PITCH_BINS, penn.NUM_TRAINING_FRAMES)
        return super().forward(frames)[:, :, None]


###############################################################################
# Utilities
###############################################################################


class Block(torch.nn.Sequential):

    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        stride,
        padding,
        length):
        layers = (
            torch.nn.ConstantPad1d(padding, 0),
            torch.nn.Conv1d(
                in_channels,
                out_channels,
                kernel_size,
                stride),
            torch.nn.ReLU())

        # Maybe add normalization
        if penn.NORMALIZATION == 'batch':
            layers += (torch.nn.BatchNorm1d(out_channels, momentum=.01),)
        elif penn.NORMALIZATION == 'instance':
            layers += (torch.nn.InstanceNorm1d(out_channels),)
        elif penn.NORMALIZATION == 'layer':
            layers += (torch.nn.LayerNorm((out_channels, length)),)
        else:
            raise ValueError(
                f'Normalization method {penn.NORMALIZATION} is not defined')

        # Add max pooling
        layers += (torch.nn.MaxPool1d(2, 2),)

        # Maybe add dropout
        if penn.DROPOUT is not None:
            layers += (torch.nn.Dropout(penn.DROPOUT),)

        super().__init__(*layers)



import math

import torch

import penn


###############################################################################
# Methods for extracting a periodicity estimate from pitch posteriorgram logits
###############################################################################


def entropy(logits):
    """Entropy-based periodicity"""
    distribution = torch.nn.functional.softmax(logits, dim=1)
    return (
        1 + 1 / math.log(penn.PITCH_BINS) * \
        (distribution * torch.log(distribution + 1e-7)).sum(dim=1))


def max(logits):
    """Periodicity as the maximum confidence"""
    if penn.LOSS == 'binary_cross_entropy':
        return torch.sigmoid(logits).max(dim=1).values
    elif penn.LOSS == 'categorical_cross_entropy':
        return torch.nn.functional.softmax(
            logits, dim=1).max(dim=1).values
    raise ValueError(f'Loss function {penn.LOSS} is not implemented')


def sum(logits):
    """Periodicity as the sum of the distribution

    This is really just for PYIN, which performs a masking of the distribution
    probabilities so that it does not always add to one.
    """
    return torch.clip(torch.exp(logits).sum(dim=1), 0, 1)



from setuptools import find_packages, setup


with open('README.md') as file:
    long_description = file.read()


setup(
    name='penn',
    description='Pitch Estimating Neural Networks (PENN)',
    version='0.0.3',
    author='Max Morrison, Caedon Hsieh, Nathan Pruyne, and Bryan Pardo',
    author_email='interactiveaudiolab@gmail.com',
    url='https://github.com/interactiveaudiolab/penn',
    install_requires=[
        'huggingface_hub', # 0.11.1
        'numpy',           # 1.23.4
        'tensorboard',     # 2.11.0
        'torch',           # 1.12.1+cu113
        'tqdm',            # 4.64.1
        'torchaudio',      # 0.12.1+cu113
        'yapecs'           # 0.0.6
    ],
    packages=find_packages(),
    package_data={'penn': ['assets/*', 'assets/*/*']},
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords=['audio', 'frequency', 'music', 'periodicity', 'pitch', 'speech'],
    classifiers=['License :: OSI Approved :: MIT License'],
    license='MIT')



from pathlib import Path


###############################################################################
# Metadata
###############################################################################


# Configuration name
CONFIG = 'fcnf0++'


###############################################################################
# Audio parameters
###############################################################################


# Width of a pitch bin
CENTS_PER_BIN = 5.  # cents

# Whether to trade quantization error for noise during inference
DITHER = False

# Minimum representable frequency
FMIN = 31.  # Hz

# Distance between adjacent frames
HOPSIZE = 80  # samples

# The size of the window used for locally normal pitch decoding
LOCAL_PITCH_WINDOW_SIZE = 19

# Pitch velocity constraint for viterbi decoding
MAX_OCTAVES_PER_SECOND = 35.92

# Whether to normalize input audio to mean zero and variance one
NORMALIZE_INPUT = False

# Number of spectrogram frequency bins
NUM_FFT = 1024

# One octave in cents
OCTAVE = 1200  # cents

# Number of pitch bins to predict
PITCH_BINS = 1440

# Audio sample rate
SAMPLE_RATE = 8000  # hz

# Size of the analysis window
WINDOW_SIZE = 1024  # samples


###############################################################################
# Directories
###############################################################################


# Location to save assets to be bundled with pip release
ASSETS_DIR = Path(__file__).parent.parent / 'assets'

# Location of preprocessed features
CACHE_DIR = Path(__file__).parent.parent.parent / 'data' / 'cache'

# Location of datasets on disk
DATA_DIR = Path(__file__).parent.parent.parent / 'data' / 'datasets'

# Location to save evaluation artifacts
EVAL_DIR = Path(__file__).parent.parent.parent / 'eval'

# Location to save training and adaptation artifacts
RUNS_DIR = Path(__file__).parent.parent.parent / 'runs'

# Location of compressed datasets on disk
SOURCE_DIR = Path(__file__).parent.parent.parent / 'data' / 'sources'


###############################################################################
# Evaluation parameters
###############################################################################


# Whether to perform benchmarking
BENCHMARK = False

# Number of steps between saving checkpoints
CHECKPOINT_INTERVAL = 25000  # steps

# List of all datasets
DATASETS = ['mdb', 'ptdb']

# Method to use for evaluation
METHOD = 'penn'

# Batch size to use for evaluation
EVALUATION_BATCH_SIZE = 2048

# Datsets to use for evaluation
EVALUATION_DATASETS = DATASETS

# Number of steps between logging to Tensorboard
LOG_INTERVAL = 2500  # steps

# Number of batches to use for validation
LOG_STEPS = 64

# Method to use for periodicity extraction
PERIODICITY = 'entropy'


###############################################################################
# Model parameters
###############################################################################


# The decoder to use for postprocessing
DECODER = 'locally_normal'

# The dropout rate. Set to None to turn off dropout.
DROPOUT = None

# The name of the model to use for training
MODEL = 'fcnf0'

# Type of model normalization
NORMALIZATION = 'layer'


###############################################################################
# Training parameters
###############################################################################


# Batch size
BATCH_SIZE = 128

# Default threshold for voiced/unvoiced classification
DEFAULT_VOICING_THRESHOLD = .1625

# Whether to stop training when validation loss stops improving
EARLY_STOPPING = False

# Stop after this number of log intervals without validation improvements
EARLY_STOPPING_STEPS = 32

# Whether to apply Gaussian blur to binary cross-entropy loss targets
GAUSSIAN_BLUR = True

# Optimizer learning rate
LEARNING_RATE = 2e-4

# Loss function
LOSS = 'categorical_cross_entropy'

# Number of training steps
STEPS = 250000

# Number of frames used during training
NUM_TRAINING_FRAMES = 1

# Number of data loading worker threads
NUM_WORKERS = 4

# Seed for all random number generators
RANDOM_SEED = 1234

# Whether to only use voiced start frames
VOICED_ONLY = False



import json
import math

import penn


###############################################################################
# Create figure
###############################################################################


def from_evaluations(names, evaluations, output_file):
    """Plot periodicity thresholds"""
    import matplotlib.pyplot as plt

    # Create plot
    figure, axis = plt.subplots(figsize=(7, 3))

    # Make pretty
    axis.spines['top'].set_visible(False)
    axis.spines['right'].set_visible(False)
    axis.spines['bottom'].set_visible(False)
    axis.spines['left'].set_visible(False)
    ticks = [0., .25, .5, .75, 1.]
    axis.set_xlim([0., 1.])
    axis.get_xaxis().set_ticks(ticks)
    axis.get_yaxis().set_ticks(ticks)
    axis.tick_params(axis=u'both', which=u'both',length=0)
    axis.set_xlabel('Unvoiced threshold')
    axis.set_ylabel('F1')
    for tick in ticks:
        axis.axhline(tick, color='gray', linestyle='--', linewidth=.8)

    # Iterate over evaluations to plot
    for name, evaluation in zip(names, evaluations):
        directory = penn.EVAL_DIR / evaluation

        # Load results
        with open(directory / 'overall.json') as file:
            results = json.load(file)['aggregate']
        with open(directory / 'periodicity.json') as file:
            optimal = json.load(file)['entropy']

        # Get thresholds and corresponding F1 values
        x, y = zip(*
            [(key, val) for key, val in results.items() if key.startswith('f1')])
        x = [float(item[3:]) for item in x] + [1]
        y = [0 if math.isnan(item) else item for item in y] + [0]

        # Plot
        line = axis.plot(x, y, label=name)
        color = line[0].get_color()
        axis.plot(optimal['threshold'], optimal['f1'], marker='*', color=color)

    # Add legend
    axis.legend(frameon=False, loc='upper right')

    # Save
    figure.savefig(output_file, bbox_inches='tight', pad_inches=0, dpi=300)



import argparse
from pathlib import Path

import penn


###############################################################################
# Periodicity threshold figure
###############################################################################


def parse_args():
    """Parse command-line arguments"""
    parser = argparse.ArgumentParser(
        description='Create periodicity threshold figure')
    parser.add_argument(
        '--names',
        required=True,
        nargs='+',
        help='Corresponding labels for each evaluation')
    parser.add_argument(
        '--evaluations',
        type=Path,
        required=True,
        nargs='+',
        help='The evaluations to plot')
    parser.add_argument(
        '--output_file',
        type=Path,
        required=True,
        help='The output jpg file')
    return parser.parse_known_args()[0]


penn.plot.threshold.from_evaluations(**vars(parse_args()))



import shutil

import huggingface_hub
import torch

import penn


###############################################################################
# Checkpoint utilities
###############################################################################


def download(repo='maxrmorrison/fcnf0-plus-plus', file='fcnf0++.pt'):
    """Download pretrained model weights from HuggingFace"""
    # Download model
    file = huggingface_hub.hf_hub_download(repo_id=repo, filename=file)

    # Make sure folder exists
    penn.DEFAULT_CHECKPOINT.parent.mkdir(exist_ok=True, parents=True)

    # Copy to checkpoint directory
    shutil.copyfile(file, penn.DEFAULT_CHECKPOINT)


def latest_path(directory, regex='*.pt'):
    """Retrieve the path to the most recent checkpoint"""
    # Retrieve checkpoint filenames
    files = list(directory.glob(regex))

    # If no matching checkpoint files, no training has occurred
    if not files:
        return

    # Retrieve latest checkpoint
    files.sort(key=lambda file: int(''.join(filter(str.isdigit, file.stem))))
    return files[-1]


def load(checkpoint_path, model, optimizer=None):
    """Load model checkpoint from file"""
    try:

        # Load checkpoint
        checkpoint_dict = torch.load(checkpoint_path, map_location='cpu')

    except FileNotFoundError as error:

        # Maybe download
        if checkpoint_path == penn.DEFAULT_CHECKPOINT:
            download()

            # Load checkpoint
            checkpoint_dict = torch.load(checkpoint_path, map_location='cpu')
        else:
            raise error

    # Restore model weights
    model.load_state_dict(checkpoint_dict['model'])

    # Restore optimizer
    if optimizer is not None:
        optimizer.load_state_dict(checkpoint_dict['optimizer'])

    # Restore training state
    step = checkpoint_dict['step']

    return model, optimizer, step


def save(model, optimizer, step, file):
    """Save training checkpoint to disk"""
    # Maybe unpack DDP
    if torch.distributed.is_initialized():
        model_state_dict = model.module.state_dict()
    else:
        model_state_dict = model.state_dict()

    # Save
    checkpoint = {
        'step': step,
        'model': model_state_dict,
        'optimizer': optimizer.state_dict()}
    torch.save(checkpoint, file)



import argparse

import penn


###############################################################################
# Preprocess datasets
###############################################################################


def parse_args():
    """Parse command-line arguments"""
    parser = argparse.ArgumentParser(description='Preprocess datasets')
    parser.add_argument(
        '--datasets',
        nargs='+',
        default=['mdb', 'ptdb'],
        help='The datasets to preprocess')
    return parser.parse_known_args()[0]


penn.data.preprocess.datasets(**vars(parse_args()))



import argparse

import penn
from pathlib import Path


###############################################################################
# Entry point
###############################################################################


def parse_args():
    """Parse command-line arguments"""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--files',
        nargs='+',
        required=True,
        type=Path,
        help='The audio files to process')
    parser.add_argument(
        '--output_prefixes',
        nargs='+',
        type=Path,
        help=(
            'The files to save pitch and periodicity without extension. '
            'Defaults to audio_files without extensions.'))
    parser.add_argument(
        '--hopsize',
        type=float,
        default=penn.HOPSIZE_SECONDS,
        help=(
            'The hopsize in seconds. '
            f'Defaults to {penn.HOPSIZE_SECONDS} seconds.'))
    parser.add_argument(
        '--fmin',
        type=float,
        default=penn.FMIN,
        help=(
            'The minimum frequency allowed in Hz. '
            f'Defaults to {penn.FMIN} Hz.'))
    parser.add_argument(
        '--fmax',
        type=float,
        default=penn.FMAX,
        help=(
            'The maximum frequency allowed in Hz. '
            f'Defaults to {penn.FMAX} Hz.'))
    parser.add_argument(
        '--checkpoint',
        type=Path,
        default=penn.DEFAULT_CHECKPOINT,
        help=(
            'The model checkpoint file. '
            f'Defaults to {penn.DEFAULT_CHECKPOINT}'))
    parser.add_argument(
        '--batch_size',
        type=int,
        default=penn.EVALUATION_BATCH_SIZE,
        help=(
            'The number of frames per batch. '
            f'Defaults to {penn.EVALUATION_BATCH_SIZE}.'))
    parser.add_argument(
        '--pad',
        action='store_true',
        help=(
            'If true, centers frames at '
            'hopsize / 2, 3 * hopsize / 2, 5 * ...'))
    parser.add_argument(
        '--interp_unvoiced_at',
        type=float,
        default=penn.DEFAULT_VOICING_THRESHOLD,
        help='Specifies voicing threshold for interpolation')
    parser.add_argument(
        '--gpu',
        type=int,
        help='The index of the gpu to perform inference on. Defaults to CPU.')

    return parser.parse_known_args()[0]


penn.from_files_to_files(**vars(parse_args()))



import torch

import penn


class Flatten(torch.nn.Module):

    def forward(self, x):
        return x.reshape(x.shape[0], -1)


class Normalize(torch.nn.Module):

    def forward(self, frames):
        return penn.normalize(frames)



import itertools
import warnings

import numpy as np
import torchaudio

import penn


###############################################################################
# Constants
###############################################################################


# MDB analysis parameters
MDB_HOPSIZE = 128  # samples
MDB_SAMPLE_RATE = 44100  # samples per second

# PTDB analysis parameters
PTDB_HOPSIZE = 160  # samples
PTDB_SAMPLE_RATE = 16000  # samples per second
PTDB_WINDOW_SIZE = 512  # samples
PTDB_HOPSIZE_SECONDS = PTDB_HOPSIZE / PTDB_SAMPLE_RATE


###############################################################################
# Preprocess datasets
###############################################################################


def datasets(datasets):
    """Preprocess datasets"""
    if 'mdb' in datasets:
        mdb()

    if 'ptdb' in datasets:
        ptdb()


###############################################################################
# Individual datasets
###############################################################################


def mdb():
    """Preprocess mdb dataset"""
    # Get audio files
    audio_files = (penn.DATA_DIR / 'mdb'/ 'audio_stems').glob('*.wav')
    audio_files = sorted([
        file for file in audio_files if not file.stem.startswith('._')])

    # Get pitch files
    pitch_files = [
        file.parent.parent /
        'annotation_stems' /
        file.with_suffix('.csv').name
        for file in audio_files]

    # Create cache
    output_directory = penn.CACHE_DIR / 'mdb'
    output_directory.mkdir(exist_ok=True, parents=True)

    # Write audio and pitch to cache
    iterator = penn.iterator(
        enumerate(zip(audio_files, pitch_files)),
        'Preprocessing mdb',
        total=len(audio_files))
    for i, (audio_file, pitch_file) in iterator:
        stem = f'{i:06d}'

        # Load and resample audio
        audio = penn.load.audio(audio_file)

        # Save as numpy array for fast memory-mapped reads
        np.save(
            output_directory / f'{stem}-audio.npy',
            audio.numpy().squeeze())

        # Save audio for listening and evaluation
        torchaudio.save(
            output_directory / f'{stem}.wav',
            audio,
            penn.SAMPLE_RATE)

        # Load pitch
        annotations = np.loadtxt(open(pitch_file), delimiter=',')
        times, pitch = annotations[:, 0], annotations[:, 1]

        # Fill unvoiced regions via linear interpolation
        pitch, voiced = interpolate_unvoiced(pitch)

        # Get target number of frames
        frames = penn.convert.samples_to_frames(audio.shape[-1])

        # Linearly interpolate to target number of frames
        new_times = penn.HOPSIZE_SECONDS * np.arange(0, frames)
        new_times += penn.HOPSIZE_SECONDS / 2.
        pitch = 2. ** np.interp(new_times, times, np.log2(pitch))

        # Linearly interpolate voiced/unvoiced tokens
        voiced = np.interp(new_times, times, voiced) > .5

        # Check shapes
        assert (
            penn.convert.samples_to_frames(audio.shape[-1]) ==
            pitch.shape[-1] ==
            voiced.shape[-1])

        # Save to cache
        np.save(output_directory / f'{stem}-pitch.npy', pitch)
        np.save(output_directory / f'{stem}-voiced.npy', voiced)


def ptdb():
    """Preprocessing ptdb dataset"""
    # Get audio files
    directory = penn.DATA_DIR / 'ptdb' / 'SPEECH DATA'
    male = (directory / 'MALE' / 'MIC').rglob('*.wav')
    female = (directory / 'FEMALE' / 'MIC').rglob('*.wav')
    audio_files = sorted(itertools.chain(male, female))

    # Get pitch files
    pitch_files = [
        file.parent.parent.parent /
        'REF' /
        file.parent.name /
        file.with_suffix('.f0').name.replace('mic', 'ref')
        for file in audio_files]

    # Create cache
    output_directory = penn.CACHE_DIR / 'ptdb'
    output_directory.mkdir(exist_ok=True, parents=True)

    # Write audio and pitch to cache
    iterator = penn.iterator(
        enumerate(zip(audio_files, pitch_files)),
        'Preprocessing ptdb',
        total=len(audio_files))
    for i, (audio_file, pitch_file) in iterator:
        stem = f'{i:06d}'

        # Load and resample to PTDB sample rate
        audio, sample_rate = torchaudio.load(audio_file)
        audio = penn.resample(audio, sample_rate, PTDB_SAMPLE_RATE)

        # Remove padding
        offset = PTDB_WINDOW_SIZE - PTDB_HOPSIZE // 2
        if (audio.shape[-1] - 2 * offset) % PTDB_HOPSIZE == 0:
            offset += PTDB_HOPSIZE // 2
        audio = audio[:, offset:-offset]

        # Resample to pitch estimation sample rate
        audio = penn.resample(audio, PTDB_SAMPLE_RATE)

        # Save as numpy array for fast memory-mapped read
        np.save(
            output_directory / f'{stem}-audio.npy',
            audio.numpy().squeeze())

        # Save audio for listening and evaluation
        torchaudio.save(
            output_directory / f'{stem}.wav',
            audio,
            penn.SAMPLE_RATE)

        # Load pitch
        pitch = np.loadtxt(open(pitch_file), delimiter=' ')[:, 0]

        # Fill unvoiced regions via linear interpolation
        pitch, voiced = interpolate_unvoiced(pitch)

        # Get target number of frames
        frames = penn.convert.samples_to_frames(audio.shape[-1])

        # Get original times
        times = PTDB_HOPSIZE_SECONDS * np.arange(0, len(pitch))
        times += PTDB_HOPSIZE_SECONDS / 2

        # Linearly interpolate to target number of frames
        new_times = penn.HOPSIZE_SECONDS * np.arange(0, frames)
        new_times += penn.HOPSIZE_SECONDS / 2.

        pitch = 2. ** np.interp(new_times, times, np.log2(pitch))

        # Linearly interpolate voiced/unvoiced tokens
        voiced = np.interp(new_times, times, voiced) > .5

        # Check shapes
        assert (
            penn.convert.samples_to_frames(audio.shape[-1]) ==
            pitch.shape[-1] ==
            voiced.shape[-1])

        # Save to cache
        np.save(output_directory / f'{stem}-pitch.npy', pitch)
        np.save(output_directory / f'{stem}-voiced.npy', voiced)


###############################################################################
# Utilities
###############################################################################


def interpolate_unvoiced(pitch):
    """Fill unvoiced regions via linear interpolation"""
    unvoiced = pitch == 0

    # Ignore warning of log setting unvoiced regions (zeros) to nan
    with warnings.catch_warnings():
        warnings.simplefilter('ignore')

        # Pitch is linear in base-2 log-space
        pitch = np.log2(pitch)

    try:

        # Interpolate
        pitch[unvoiced] = np.interp(
            np.where(unvoiced)[0],
            np.where(~unvoiced)[0],
            pitch[~unvoiced])

    except ValueError:

        # Allow all unvoiced
        pass

    return 2 ** pitch, ~unvoiced



import torch

import penn


###############################################################################
# DeepF0 model implementation
###############################################################################


class Deepf0(torch.nn.Sequential):

    def __init__(self, channels=128, kernel_size=64):
        layers = (penn.model.Normalize(),) if penn.NORMALIZE_INPUT else ()
        layers += (
            CausalConv1d(1, channels, kernel_size),
            Block(channels, channels, kernel_size, 1),
            Block(channels, channels, kernel_size, 2),
            Block(channels, channels, kernel_size, 4),
            Block(channels, channels, kernel_size, 8),
            torch.nn.AvgPool1d(kernel_size),
            penn.model.Flatten(),
            torch.nn.Linear(2048, penn.PITCH_BINS))
        super().__init__(*layers)

    def forward(self, frames):
        # shape=(batch, 1, penn.WINDOW_SIZE) =>
        # shape=(batch, penn.PITCH_BINS, penn.NUM_TRAINING_FRAMES)
        return super().forward(frames)[:, :, None]


###############################################################################
# Utilities
###############################################################################


class Block(torch.nn.Sequential):

    def __init__(
        self,
        input_channels,
        output_channels,
        kernel_size,
        dilation):
        if penn.NORMALIZATION == 'weight':
            norm_conv = [
                torch.nn.utils.weight_norm(
                    torch.nn.Conv1d(output_channels, output_channels, 1))]
        elif penn.NORMALIZATION == 'layer':
            norm_conv = [
                torch.nn.LayerNorm(
                    (output_channels, penn.NUM_TRAINING_SAMPLES)),
                torch.nn.Conv1d(output_channels, output_channels, 1)]
        else:
            raise ValueError(
                f'Normalization method {penn.NORMALIZATION} is not defined')

        super().__init__(*(
            [
                CausalConv1d(
                    input_channels,
                    output_channels,
                    kernel_size,
                    dilation=dilation),
                torch.nn.ReLU()
            ] + norm_conv))

    def forward(self, x):
        return torch.nn.functional.relu(super().forward(x) + x)


class CausalConv1d(torch.nn.Conv1d):

    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        stride=1,
        dilation=1,
        groups=1,
        bias=True):
        self.pad = (kernel_size - 1) * dilation
        super().__init__(
            in_channels,
            out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=self.pad,
            dilation=dilation,
            groups=groups,
            bias=bias)

    def forward(self, input):
        result = super().forward(input)
        if self.pad != 0:
            return result[:, :, :-self.pad]
        return result
